/**
 * This type of task supports a rudimentary state calling convention.
 *
 * The task can change states by calling other states. Those states
 * can then return to the calling state. Parameters have not been
 * implemented but a subclass could easily do so.
 *
 * @version 1.0 8/25/02
 *
 * @author Parallax, Inc.
 */

package stamp.util.os;

import stamp.util.*;

public abstract class CallableTask extends InterruptTask {
  // stack is exclusive to task
  protected CallStackEntry stack = null ;

  /**
   * Get current parameter
   *
   * @return parameter passed in latest callState
   */
  public Object getParameter () {
    return ( stack == null ) ? null : stack.parameter ;
  }

  /**
   * Call a state.
   *
   * @param callState next state to enter
   * @param calling parameter
   * @param returnState state to enter when <code>returnState()</code> is called
   */
  public void callState ( int callState, Object parameter, int returnState ) {
    // Add a new stack entry
    stack = CallStackEntry.getEntry ( returnState, stack, parameter ) ;

    // Set new state
    nextState ( callState ) ;
  }

  /**
   * Call a state.
   *
   * @param callState next state to enter
   * @param returnState state to enter when <code>returnState()</code> is called
   */
  public void callState ( int callState, int returnState ) {
    callState ( callState, null, returnState ) ;
  }


  /**
   * Return to prior state specified in the last <code>callState</code>.
   * Terminate task if return performed without prior call.
   */
  public void returnState () {
    CallStackEntry oldEntry ;

    if ( stack == null ) {
      state = terminate ;
    } else {
      // Setup return state
      state = stack.returnState ;

      // Pop the stack
      oldEntry = stack ;
      stack = oldEntry.nextEntry ;

      // Free entry
      oldEntry.free () ;

      // Note: The following should be added if garbage collection is supported
      // stackEntry.parameter = null
    }
  }
}