package stamp.util.os;
import stamp.core.*;

/**
 * This class extends the SimpleTask class to check a button, switch or input signal.
 * It memorises pressed and released states between status reads.
 * (for auto repeat and debounce use the Button class.)
 * This class assumes a 'long' time between status reads so debounce is not necessary.
 *
 * Four possible button/switch connections:
 *
 * stamp pin o-----+----[10k]---- VCC     stamp pin o----+----[10k]---- GND
 *                 |                                     |
 *                  /                                     /
 *                 /  NO to GND                          /  NO to VCC
 *                 |                                     |
 *                GND                                   VCC
 *
 * stamp pin o-----+----[10k]---- VCC     stamp pin o----+----[10k]---- GND
 *                 |                                     |
 *                 +                                     +
 *                 +  NC to GND                          +  NC to VCC
 *                 |                                     |
 *                GND                                   VCC
 *
 * By passing this object to classes that have methods that require a significant
 * amount of time, such methods can call execute() to keep performing this task.
 *
 * Revision History:
 * 02/07/2005: v1.0: Initial release.
 * </pre><p>
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * version 1.0 December 7, 2005
 */

public class CheckPin extends SimpleTask {

  public static final int UP = 0;
  public static final int DOWN = 1;

  public static final int PRESSED_DOWN = 3;
  public static final int RELEASED_UP = 4;

  public static final int PRESSED_RELEASED_UP = 6;
  public static final int RELEASED_PRESSED_DOWN = 7;

  private int pin; //pin button connects to
  private boolean upState; //true for default high input, false for default low input
  private boolean previous=false, current=false; //button states
  private int status = UP;

  /**
   * Constructor.
   *
   * @param pin Javelin I/O pin button/switch/signal connects to.
   * @param upState True for default high input, false for default low input.
   *                button type   switches to   upState
   *                    NO            GND        true
   *                    NO            VCC        false
   *                    NC            GND        false
   *                    NC            VCC        true
   */
  public CheckPin(int pin, boolean upState) {
    super();
    this.pin = pin;
    this.upState = upState;
  }

  /**
   * Get pin input status
   *
   * @return Status value.
   *           bit0: current button state (0 = up, 1 = down)
   *           bit1: 0 if button not pressed since previous status read, 1 if pressed since previous status read
   *           bit2: 0 if button not released since previous status read, 1 if released since previous status read
   *           Internal pressed and released flags (bits 1 and 2) are cleared upon return.
   *
   *         Possible status values (also defined public):
   *         UP (0)                    - button was up on previous status read and is still up
   *         DOWN (1)                  - button was down on previous status read and is still down
   *         PRESSED_DOWN (3)          - button was up on previous status read and is now down
   *         RELEASED_UP (4)           - button was down on previous status read and is now up
   *         PRESSED_RELEASED_UP (6)   - button has been pressed and released since previous status read and is now up
   *         RELEASED_PRESSED_DOWN (7) - button has been released and pressed since previous status read and is now down
   */
  public int status() {
    int s = status;
    status &= 0x01; //clear pressed and released flags
    return s;
  }

  /**
   * Execute simple task.
   */
  public void execute() {
    previous = current;
    current = (upState) ? !CPU.readPin(pin) : CPU.readPin(pin);
    status &= 0x06;         //button currently up
    if (current) status++;  //button currently down
    if (!previous && current) { //button pressed
      status |= 0x02;   //flag is cleared by status()
    }
    if (previous && !current) { //button released
      status |= 0x04;   //flag is cleared by status()
    }
  }

}