/*
 * Semaphore class
 * <p>
 * Used by a Task to control a resource. It can be acquired
 * by one Task at a time. Subsequent tasks will be suspended
 * and added to the Semaphore list
 *
 * @author Parallax, Inc.
 *
 * @version 2.1 2/14/03   support Task.taskStatus
 * @version 2.0 8/30/02   extends Event. Added semaphoreList
 * @version 1.1 7/24/02   acquire() uses TaskManager.currentTask
 * @version 1.0 7/24/02   initial version
 */

package stamp.util.os;

import stamp.util.*;

/**
 * Standard semaphore support for Task object.
 * It starts in the ready state.
 */

public class Semaphore extends Event {
  protected static Semaphore semaphoreList ;

  protected Semaphore nextSemaphore ;
  protected Task acquiredTask ;
  protected Task waitingTaskList ;

  /**
   * Create semaphore object.
   */
  public Semaphore () {
    nextSemaphore = semaphoreList ;
    semaphoreList = this ;
    acquiredTask = null ;
    waitingTaskList = null ;
  }

  /**
   * See if semaphore is ready
   *
   * @return true if semaphore is ready to be acquired
   */
  public boolean ready () {
    return ( acquiredTask == null ) ;
  }

  /**
   * Acquire semaphore. Sets task's next state.
   *
   * The task should exit execute() for this to take affect.
   * Use ready() if the task must conditionally acquire the semaphore.
   * Does a task.suspend() if semaphore is not ready.
   * If the semaphore is acquired then execution can continue. This allows
   * the task to release the semaphore immediately after performing
   * some action if possible. Set the nextState appropriately.
   * Remember not to do more than one matching release().
   * <p>
   * Example:
   * <code><pre>
   *  case 9: // this always yields control
   *     semaphore.acquire(11);
   *     break;
   *  case 10: // yields control if not acquired
   *     if ( semaphore.acquire(11))
   *       break;
   *  case 11:
   *     // do something
   *     semaphore.relese() ;
   *     nextState(12);
   *     break ;
   * </pre></code>
   *
   * @return <code>false</code> if semaphore acquired, <code>true</code> if suspended
   */
  public boolean acquire ( int nextState ) {
    Task currentTask = Task.currentTask ;

    // force next state
    currentTask.nextState ( nextState ) ;

    if ( ready ()) {
      acquiredTask = Task.currentTask ;
      return false ;
    }

    // remove task from run list
    if ( ! currentTask.suspend ()) {
      return true ; // task was already suspended
    }

    // Indicate task is waiting on semaphore
    currentTask.taskStatus = Task.taskSemaphoreWait ;

    // add task to waiting list
    if ( waitingTaskList == null ) {
      // list was empty
      waitingTaskList = Task.currentTask ;
    } else {
      // add to end of list
      // Note: task.nextTask should be null
      Task lastTask = waitingTaskList ;

      // walk to end of list
      while ( lastTask.nextTask != null )
        lastTask = lastTask.nextTask ;

      // append task to end of list
      lastTask.nextTask = Task.currentTask ;
      Task.currentTask.nextTask = null ; // just in case
    }

    return true ;
  }

  /**
   * Release semaphore. Start next task if one is waiting.
   * Assumes the task that is doing the release is allowed to do so.
   */
  public void release () {
    if ( waitingTaskList == null ) {
      acquiredTask = null ;
    }
    else {
      // Tasks are suspended with the nextState set.
      // Resume the first waiting task.
      acquiredTask = waitingTaskList ;

      waitingTaskList = waitingTaskList.nextTask ;
      acquiredTask.nextTask = null ;

      // taskStatus must be taskSuspended for resume to work
      acquiredTask.taskStatus = Task.taskSuspended ;
      acquiredTask.resume () ;
    }
  }
  /**
   * Release semaphore when an event occurs.
   * The Semaphore class is based on the Event class
   * that has only the notify method.
   */
  public void notify (Object object) {
    release () ;
  }

  /**
   * Get name.
   *
   * @return name of semaphore.
   */
  public String name () {
    return "Semaphore" ;
  }
}