/*
 * Co-operative, multitasking state machine operating system
 * <p>
 * This very basic operating system using round robin task scheduling.
 * The task list is a circular linked list for efficiency reasons.
 *
 * Note: If the TaskStatus class is not used and you need a few more
 * bytes then it is possible to eliminate taskStatusList and
 * nextTaskStatus along with references in the constructor.
 *
 * @author William Wong
 *
 * @version 3.1 2/14/02 added taskStatus and related methods
 *                      removed running()
 * @version 3.0 8/23/02 eliminated execute() parameters and TaskManager
 * @version 2.0 7/23/02 state variable to execute()
 * @version 1.0 7/21/02
 */

package stamp.util.os;

import stamp.util.*;
import stamp.core.*;

/**
 * A task that is run whenever there are free cycles.
 * <p>
 * Note: The nice way for an external task to terminate a task is
 * to use taskToAbort.nextState(abort). The alternative is to
 * use taskToAbort.stop() but this does not allow any cleanup.
 */

 public abstract class Task extends Event {
  /**
   * taskState() result
   */
  final public static int taskRunning       = 0 ;

  /**
   * taskState() result
   */
  final public static int taskSuspended      = 1 ;

  /**
   * taskState() result
   */
  final public static int taskSemaphoreWait = 2 ;

  /**
   * Initial value of state in execute()
   */
  final public static int initialState = 0 ;
  final public static int checkTimer   = -1 ; // for sleep() support
  final public static int interrupt    = -2 ; // see InterruptTask
  final public static int terminate    = -3 ;
  final public static int stopped      = -4 ;

  /**
   * Timer object to handle task sleep() requests
   */
  public    Timer timer    = new Timer () ;

  /**
   * Next active task in circular list
   */
  protected static Task taskStatusList  = null ;

  /**
   * Linked list for taskStatusList
   */
  protected Task nextTaskStatus  = null ;

  /**
   * Value returned by taskState()
   */
  protected int taskStatus ;

  /**
   * Next active task in circular list
   */
  protected Task nextTask  = null ;

  /**
   * State variable that can be used in execute()
   */
  protected int state      = initialState ;

  /**
   * Next state after timeout occurs
   */
  protected int sleepState = checkTimer ;

  /**
   * Timout values
   */
  protected int hi, lo ;

  /**
   * Setup and start task.
   */
  public Task() {
    addTask ( this ) ;

    // Add task to status list
    nextTaskStatus = taskStatusList ;
    taskStatusList = this ;
  }

  /**
   * Get task name. Default is an empty string.
   *
   * @return name of task.
   */
  public String name () {
    return "Task" ;
  }

  /**
   * Get task status
   *
   * @returns task status (taskRunning, taskSuspended, taskSemaphoreWait)
   */
  public int taskStatus() {
    return taskStatus;
  }

  /**
   * Get current task state
   *
   * @return task state
   */
  public int getState () {
    return state ;
  }

  /**
   * Remove task from task manager's task list if task is running.
   * This can be called by any task.
   *
   * @return true if task was removed from the list
   */
  public boolean suspend () {
    if ( taskStatus == taskRunning ) {
      // Note: myManager should not be null if nextTask is not null
      taskStatus = taskSuspended ;
      return removeTask ( this ) ;
    }

    return false ;
  }

  /**
   * Suspend and set next state to stopped.
   *
   */
  public boolean stop() {
    if ( suspend ()) {
      state = stopped ;
      return true ;
    }

    return false ;
  }

  /**
   * Add task back into task manager's task list if task is not running
   * This can be called by any task but the task must have a manager.
   *
   */
  public boolean resume () {
    if ( taskStatus == taskSuspended ) {
      // Note: myManager should not be null if nextTask is not null
      addTask ( this ) ;
      return true;
    }
    return false ;
  }

  /**
   * Same as resume()
   */
  public boolean start() {
    return resume () ;
  }

  /**
   * Allows the task to be used as an Event value.
   */
  public void notify( Object object ) {
    resume () ;
  }

  /**
   * Set the state for the next execute() call.
   *
   * @param state next task state value
   */
  public void nextState(int state) {
    this.state = state ;
  }

  /**
   * Set the state for the next execute() call.
   * Suspend the task. It must be resumed by another task.
   *
   * @param state next task state value
   */
  public void sleep(int state) {
    this.state = state ;
    suspend();
  }

  /**
   * Setup to poll timer and continue with the next state after
   * the elapsed number of ticks occurs.
   *
   * @param nextState state to enter when the timeout occurs
   * @param hi hi timeout value, see Timer.timeout(hi,lo) for details
   * @param lo lo timeout value
   */
  public void sleep ( int nextState, int hi, int lo ) {
    this.hi = hi ;
    this.lo = lo ;
    timer.mark () ;
    sleepState = nextState ;
    state = checkTimer ;
  }

  /**
   * Setup to poll timer and continue with the next state after
   * the elapsed number of ticks occurs.
   *
   * @param nextState state to enter when the timeout occurs
   * @param msec timeout value in milliseconds
   */
  public void sleep ( int nextState, int timeMS ) {
    sleep ( nextState, timeMS/569, (timeMS%569)*115) ;
  }

  /**
   * Setup to poll timer and continue with the next state after
   * the elapsed number of ticks occurs.
   *
   * @param nextState state to enter when the timeout occurs
   * @param sec timeout value in seconds
   */
  public void sleepSec ( int nextState, int timeS ) {
    sleep ( nextState, (timeS*18)/10, 0);
  }


  /**
   * Perform a period task. All subclasses must implement this method.
   * An idle task is executed whenever the GUI is idle. Implementations of
   * <code>execute()</code> should not take more than 100ms to ensure that the
   * interface remains responsive.
   */
  protected abstract void execute() ;


  // Task manager support follows

  /**
   * Class variable: current task in task list
   */
  static protected Task taskList = null;

  /**
   * Current active task.
   */
  static protected Task currentTask = null ;

  /**
   * Get currently executing task object.
   *
   * @returns currently executing task
   */
  static public Task getCurrentTask () {
    return currentTask ;
  }

  /**
   * Add a task to task list
   *
   * @param aNewTask task to be added to task list
   */
  public static void addTask(Task aNewTask) {
    if ( taskList == null ) {
      aNewTask.nextTask = aNewTask ;
    }
    else
    {
      aNewTask.nextTask    = taskList.nextTask ;
      taskList.nextTask = aNewTask ;
    }

    aNewTask.taskStatus = taskRunning ;
    taskList = aNewTask ;
  }

  /**
   * Remove a task from the task list
   *
   * @param aNewTask task to be added to task list
   *
   * @return true if task removed from the list
   */
  public static boolean removeTask(Task aTaskToRemove) {
    // handle empty list first

    if ( taskList != null ) {
      // scan list from current task and remove it if found

      Task scanTask = taskList ;

      do {
        if ( scanTask.nextTask == aTaskToRemove ) {
          // found task to remove

          if ( scanTask == aTaskToRemove ) {
            // removing the only task in the list
            taskList = null ;
          }
          else
          {
            // check if current task being removed

            if ( taskList == aTaskToRemove ) {
              taskList = aTaskToRemove.nextTask ;
            }

            scanTask.nextTask = aTaskToRemove.nextTask ;
          }

          aTaskToRemove.nextTask  = null ;
          return true ;
        }

        // check next task in the list
        scanTask = scanTask.nextTask ;
      }
      while ( scanTask != taskList ) ; // exit if full list scanned
    }

    return false ;
  }

  /**
   * This method will run forever executing each task in the readyTask list
   * in a round robin fashion. It returns when there are no running tasks.
   */
  public static void TaskManager() {
    while ( taskList != null ) {
      // change taskList so execute() can change it via removeTask
      currentTask = taskList ;
      taskList = taskList.nextTask ;

      if (currentTask.state == checkTimer) {
        // Task is waiting for a timeout
        if (currentTask.timer.timeout(currentTask.hi,currentTask.lo)) {
          // Timeout occurred. Set next state.
          currentTask.state = currentTask.sleepState ;

          // Abort if the next state was set improperly
          if ( currentTask.state == checkTimer ) {
            currentTask.state = terminate ;
          }
        }
      } else {
        // Just call the next state
        currentTask.execute () ;
      }
    }
  }
}