/**
 * Tone generator for use with TaskManager
 * <p>
 * Generates tones for a specified period using the TaskTimer support.
 * It does not use CPU.delay like FREQOUT does.
 *
 * @version 1.0 7/23/02
 * @author Parallax, Inc.
 */

package stamp.util.os;

import stamp.core.*;


public class TaskToneGenerator extends TimerTask {
  protected PWM pwm ;
  protected int pin ;
  protected int tones[] ;
  protected int tonesIndex ;
  protected boolean pwmRunning = false ;

  final static int endTone = 1 ;  // state when done done

  final static int noMoreTones = -1 ;

  /**
   * Setup the tone generator output.
   *
   * @param pin set output pin such as CPU.pin10
   */
  public TaskToneGenerator ( int pin ) {
    this.pin = pin ;
    pwm = new PWM ( pin ) ;
  }

  /**
   * Get task name.
   *
   * @return name of task.
   */
  public String name () {
    return "TaskToneGenerator" ;
  }

  /**
   * Set output frequency
   *
   * @param frequency frequency in hertz (0 - 32k)
   */
  protected void setFrequency ( int frequency ) {
    if ( frequency == 0 )
      frequency = 1 ;

    int halfCycleTime = 12000 / frequency ;

    pwm.update ( halfCycleTime, halfCycleTime ) ;
  }

  /**
   * Play a tone for a fixed amount of time.
   *
   * @param frequency frequency in hertz (0 - 32k)
   * @param time duration in milliseconds
   */
  public void playTone (int frequency, int time) {
    tonesIndex = noMoreTones ;
    playToneNow ( frequency, time ) ;
  }

  /**
   * Play a set of tones. The array contains an even number of items.
   * The first value in an item is the frequency.
   * The second is the duration.
   * Does nothing if there are not at least 2 elements in the array.
   *
   * @param tones tone array with (frequency, duration) pairs
   */
  public void playTones ( int tones[] ) {
    if ( tones.length >= 2 )
    {
      this.tones = tones ;
      tonesIndex = 0 ;
      startNextTone () ;
    }
  }

  /* Internal routine */
  protected void playToneNow ( int frequency, int time ) {
    // Setup PWM
    setFrequency ( frequency ) ;

    // Check if PWM and task already running
    if ( ! pwmRunning ) {
      pwmRunning = true ;
      pwm.start () ;    // start PWM
    }

    /* Sleep for the specified amount of time.
     * Must be after timerRunning check performed above
     * because the following sets the timerRunning status
     */
    sleep(time) ;
  }


  /* Internal routine  */
  protected void startNextTone () {
    // tonesIndex will always reference a valid pair

    playToneNow ( tones[tonesIndex], tones[tonesIndex+1] ) ;
    tonesIndex += 2 ;

    // Adjust tonesIndex if this was the last tone in the list
    if (( tonesIndex + 2 ) > tones.length ) {
      tonesIndex = noMoreTones ;
    }
  }

  /**
   * Turn off tone generation if tone is being played
   */
  public void stopTone () {
    interrupt ( 0 ) ;
  }

  /**
   * Turn off tone generation if it is on.
   *
   * @param interruptValue ignored
   *
   * @return <code>false</code>, stop task
   */
  public boolean handleInterrupt ( int interruptValue ) {
    if ( pwmRunning ) {
      tonesIndex = noMoreTones ;    // disable on next timeout
      handleTimeout () ;            // force timeout
    }
    return false ;
  }

  /**
   * Task execution routine called by TaskManager
   *
   * @return <code>false</code> if done, <code>true</code> if more tones
   */
  public boolean handleTimeout () {
    if ( tonesIndex == noMoreTones ) {
      // No more tones. Turn off PWM
      pwmRunning = false ;
      pwm.stop () ;      // turn off tones
      CPU.readPin (pin);
      return false ;     // stop task
    } else {
      // Get next tone
      startNextTone () ; // startup timer and next tone
      return true ;      // continue task
    }
  }
}