/**
 * Task for simplified timer-based operations
 * <p>
 * This type of task is designed to be started when the timer is
 * started. The <code>handleTimeout</code> method is called
 * when the timeout occurs. The timer can be restarted or the
 * task can be terminated.
 *<p>
 * A typical use is a tone generator that will be used with the
 * TaskManager. This allows other tasks to run while the timer
 * is running. The <code>handleTimeout</code> method can also
 * be defined. By default, calling the <code>interrupt</code>
 * method will terminate the timer.
 *
 * @version 1.0 7/23/02
 * @author William Wong
 */

package stamp.util.os;


public abstract class TimerTask extends InterruptTask {
  final static int timeout = 1 ;

  /**
   * Called when interrupt invoked.
   * Redefine if subclass will handle this state.
   *
   * @return <code>true</code> if resume, <code>false</code> if terminate
   */
  public boolean handleInterrupt ( int interruptValue ) {
    return false ;
  }

  /**
   * Called when timeout occurs.
   * Redefine if subclass will handle this state.
   *
   * @return <code>true</code> if resume, <code>false</code> if terminate
   */
  public abstract boolean handleTimeout () ;

  /**
   * Check if timer is running
   *
   * @return <code>true</code> if timer is running
   */
  public boolean timerRunning () {
    return state == checkTimer ;
  }

  /**
   * Task execution routine called by TaskManager
   *
   * @param state current task state
   */
  public void execute () {
    switch ( state ) {
    case interrupt:
      if ( handleInterrupt ( interruptValue ))
        resumeInterrupt () ;
      else
        nextState ( terminate ) ;
      break ;

    case timeout:
      if ( handleTimeout ())
        break ;     // continue if timeout handled

      // otherwise fall through to stop() the task

    // case initialState:
    // case terminate:
    default:
      stop () ;
      break ;
    }
  }

  /**
   * Start timer and call handleTimeout() on timeout.
   *
   * @param hi hi timeout value, see Timer.timeout(hi,lo) for details
   * @param lo lo timeout value
   */
  public void sleep ( int hi, int lo ) {
    sleep ( timeout, hi, lo ) ;
    resume () ;
  }

  /**
   * Start timer and call handleTimeout() on timeout.
   *
   * @param msec timeout value in milliseconds
   */
  public void sleep ( int timeMS ) {
    // use super or you get the local sleep method
    super.sleep ( timeout, timeMS ) ;
    resume () ;
  }

  /**
   * Start timer and call handleTimeout() on timeout.
   *
   * @param sec timeout value in seconds
   */
  public void sleepSec ( int timeS ) {
    sleepSec ( timeout, timeS );
    resume () ;
  }
}