{{
@file pjMemory.spin
Java Heap Memory Manager
}}

{{
  Heap chunks are allocated from heap start down (decrementing list).
  Each chunk length is a power of 2. The minimum chunk size is MINCHUNKSIZE.
  The new entry is the heaphead so it is the first accessed.
  The address of the new entry is the *next* address of the next new entry.
  The data for the new entry is the address + 2 for word wide addresses.
  An entry is free if bit 0 of the next pointer is clear, else it is used.
  Entries are word aligned (value & -2 = value & $fffe or value & $ffffffe).
}}

  MINCHUNKSIZE = $10            'minimum heap allocation chunk size
  HEADLENGTH = 2                'heap list header length in bytes
  
  GCEPOCHCOUNT = $20             '$10 '$40 Garbage Collected every GC EPOCH

{
con
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

pub main | ptr, n, v
{{
Initialize heap and test private alloc/free
}}
  sx.start(31,30,0,115200)
  waitcnt(clkfreq*1+cnt)
  sx.out(0)
  
  init($7f04)
  palloc(7)
  ptr := palloc(7)
  pfree(ptr)
  palloc(7)
  palloc(7)
  
  repeat n from 0 to 511
    if ptr
      pfree(ptr)
    ptr := palloc((v?) & $7ff)
    ifnot ptr
      quit
    'show
  repeat
'}  

var
  long heaphead

dat
  gcCount long 0
  gcEpoch long GCEPOCHCOUNT     ' garbage is collected every gcEpoch allocs

obj
  pjd : "pjData"
#ifdef FDX
  sx  : "FullDuplexSingleton"
#else
  sx  : "SingleSerial"
#endif

pub init(headp) | start, end
  heaphead := headp-4
  heaphead &= -2 
  setNext(heaphead,0)
{
  sx.str(string($d,$d,"Alloc Init $"))
  sx.hex(heaphead, 4)
  sx.out($d)
'}

pub alloc(num) | dataptr, sp, stkptr
'' user alloc
'' allocates a block of num bytes memory in a list
'' saves block pointer to Java HeapPointer
''
  collectGarbage                ' moved to PASM runner

  dataptr := palloc(num)
  ifnot dataptr                 ' forceGC now done in PASM ... reenable here is a waste
    sx.str(string("Out of Memory!",$d))
    show
    repeat
      
  pjd.setAP(dataptr)
  if dataptr =< pjd.getHP
    pjd.setHP(getLinkPtr(dataptr))

  'show
  return dataptr

  {
pub free(ptr)
'' user free memory
'' returns zero on error
''
  return pfree(ptr)
}

'{
'GarbageCollection done in PASM runner now.
pub collectGarbage
' garbage collection service
  ' do garbage collection once every gcEpoch
  if ++gcCount > gcEpoch
    gcCount := 0
    forceGC
'}
    
pri getDataPtr(linkptr)
  return linkptr+2

pri getLinkPtr(dataptr)
  return dataptr-2
  
pri isFree(linkPtr)
  return !getUsed(linkPtr) & 1

pri setFree(linkPtr)
  if linkPtr
    word[linkPtr] &= -2

pri setUsed(linkPtr)
  if linkPtr
    word[linkPtr] |= 1

pri getUsed(linkPtr)
  ifnot linkPtr
    return 0
  return word[linkPtr] & 1

pri getNext(linkPtr)
  if linkPtr
    return word[linkPtr] & -2
  return 0

pri setNext(linkPtr, nextlinkPtr)
  if linkPtr
    word[linkPtr] := nextlinkPtr

pri getSize(linkPtr)
  if linkPtr
    return (getNext(linkPtr)-linkPtr)
  return 0

pri getDataSize(linkPtr)
  if linkPtr
    return (getNext(linkPtr)-linkPtr-headlength)
  return 0

pri chunksize(num) | tmp
  result := num
  num := >| num
  result := 1 << (num)
  if result < MINCHUNKSIZE
    result := MINCHUNKSIZE

pri additem(headptr, num) | newptr
'' add new pointer to list
  'debug(string("additem(length):"),num)
  num := chunksize(num + headlength)
  newptr :=(headptr - num)      ' always enough space to make even
  newptr &= -2                  ' even address
  setNext(newptr, headptr)      ' headptr->next end->next = new pointer
  setUsed(newptr)               ' entry is used
  headptr := newptr             ' nextptr is now headptr
  'debug(string("new dataptr"),headptr)
  'debug(string("new dataptr.next"),getNext(headptr))
  'debug(string("new dataptr.length:"),getNext(headptr))
  return headptr 

pri insert(prvptr, num) | newptr
'' special case of add
'' insert new pointer to list

  num := chunksize(num + headlength)

  newptr :=(prvptr + num)       ' always enough space to make even
  newptr &= -2                  ' even address
  
  setUsed(prvptr)               ' entry is used regardless of insert/replace

  if newptr <> getNext(prvptr)  ' if prev free block is split, keep leftovers free
    'debug(string("inserted(length):"),num)
    setNext(newptr, getNext(prvptr)) ' newptr.next = prvptr.next
    setFree(newptr)             ' new ptr is free
    setNext(prvptr, newptr)     ' prvptr->next end->next = new pointer
  'else
    'debug(string("recycled(length):"),num)
    
{
  debug(string("prvptr"),prvptr) 
  debug(string("prvptr.next"),getNext(prvptr)) 
  debug(string("prvptr.used"),getUsed(prvptr))
  debug(string("newptr"),newptr) 
  debug(string("newptr.next"),getNext(newptr)) 
  debug(string("newptr.used"),getUsed(newptr))
'}
  return prvptr

pri palloc(numbytes) | datptr, nxtptr, prvptr, tmpptr, len
'' allocate memory using a linked list
''  
' should we survey the unused links and reallocate them?
'    
  nxtptr := heaphead
  prvptr := 0

'  longfill(@bstats,0,BSTATLEN)
  
  repeat while getNext(nxtptr)

    'debug(string("  req  length:"),chunksize(numbytes+HEADLENGTH))
    'debug(string("  nxtptr"),nxtptr)

    if prvptr
{    
      len := getSize(prvptr)
      debug(string("  prev length:"),len)
}
      if isFree(prvptr) and (chunksize(numbytes+HEADLENGTH) == getSize(prvptr))
        prvptr := insert(prvptr, numbytes)
{        
        debug(string("  new dataptr"),getDataPtr(prvptr)) 
        len := getSize(prvptr)
        debug(string("  new data length:"),len)
}        
        return getDataPtr(prvptr)

    prvptr := nxtptr
    nxtptr := getNext(nxtptr)

'  showFreeStats                 ' show only if garbage not recycled

  ' if no insert, additem to end of list
  heaphead := additem(heaphead, numbytes)
  return getDataPtr(heaphead)
{
pri pfree(ptr) | nxtptr
  'debug(string("Free"),getLinkPtr(ptr))
  'debug(string(" Size"),getSize(getLinkPtr(ptr)))
  nxtptr := heaphead
  repeat while getNext(nxtptr)
    if nxtptr == getLinkPtr(ptr)
      if getUsed(nxtptr)
        setFree(nxtptr)
      return true
    nxtptr := getNext(nxtptr)

  return false
}

{
pri showFreeStats | n
  sx.str(string($d," free stats: "))
  repeat n from 0 to BSTATLEN-1
    'sx.hex(n, 2) ' n 0 keeps $10 count, n1 keeps $20 count, n3 keeps $80 count ....
    'sx.out(":")
    sx.hex(bstats[n], 2)
    sx.out(" ")
'}
    
pri debug(name, data)
{
  sx.str(name)
  sx.out(" ")
  sx.hex(data,4)
  sx.out($d)
'}
    
'{
pub show | nextp, len, top
  sx.str(string("Show Objects:",$d))
  debug(string("headptr"),heaphead)
  len := 0
  nextp := heaphead             ' print used list
  repeat while getNext(nextp)   ' never print top of heap entry
  '{
    if(getUsed(nextp) <> 0)
      sx.str(string(" Used $"))
    if(isFree(nextp) <> 0)
      sx.str(string(" Free $"))
    sx.hex(getDataPtr(nextp),4)
    sx.str(string(" Size $"))
    sx.hex(getSize(nextp),4)
    sx.str(string(" @ $"))
    sx.hex(nextp,4)
    sx.str(string(" -> $"))
    sx.hex(getNext(nextp),4)
    sx.out($d)
  '}    
    nextp := getNext(nextp)
    len++
  sx.str(string(" Object Count: "))
  sx.dec(len)
  sx.out($d)
{
  len := 0
  top := pjd.getHB-1
  repeat nextp from heaphead to top step 2
    ifnot len++ // 8
      sx.out($d)
      sx.hex(nextp,4)
      sx.out(":")
    sx.out(" ")
    sx.hex(word[nextp],4)
  sx.out($d)
'}
'}

{
pri printStack | sp, stkptr
  stkptr := pjd.getSP
  repeat sp from pjd.getBP to stkptr-2 step 2
    pstack(sp)

pri pstack(sp)
  result := pjd.getProgOffset
  sx.str(string("stack "))
  sx.hex(sp+result,4)
  sx.out(" ")
  sx.hex(swap16(word[sp+result]),4)
  sx.out($d)

pri pstatic(sp)
  sx.str(string("static "))
  sx.hex(sp,4)
  sx.out(" ")
  sx.hex(word[sp],4)
  sx.out($d)

pri pheap(sp)
  sx.str(string("heap "))
  sx.hex(sp,4)
  sx.out(" ")
  sx.hex(word[sp],4)
  sx.out($d)

pri printheap | hp, heaptop
  heaptop := pjd.getHS
  repeat hp from heaphead to heaptop step 2
    if word[hp] => heaphead
      pheap(hp)

'}

'{
pri swap16(val)
  result := val & $ff
  val >>= 8
  result <<= 8
  result |= val

pri forceGC | sp, stkptr, heapptr, heaptop, nxtptr, inuse, progptr, statptr, hp, glen, nlen
'' force garbage collection
'' move garbage collection to pasmrunner if possible

  progptr := pjd.getProgOffset
  stkptr := pjd.getSP
  nxtptr := heaphead
  statptr := pjd.getHeapBase
  heaptop := pjd.getHS

{
  debug(string("progptr"),progptr)
  debug(string("heaphead"),heaphead)
  debug(string("heaptop"),heaptop)
  debug(string("statptr"),statptr)
}
'  debug(string("forceGC"),heaphead)

  ' foreach heap item, check for stack entry
  ' if no stack entry, mark heap item free
  '
  glen := 0
  nlen := 0
  
  repeat while getNext(nxtptr)

    inuse := 0
    heapptr := getDataPtr(nxtptr)
    'debug(string("Looking for HeapPtr"),heapptr)

    ' search static vars
    ifnot inuse
      repeat sp from heaptop to statptr step 2
        'pstatic(sp)
        if heapptr == word[sp]
          inuse++
          'debug(string("HeapPtr in static"),sp)
          quit

    ' search stack
    ifnot inuse
      repeat sp from pjd.getBP to stkptr step 2
        'pstack(sp)
        if heapptr == swap16(word[sp+progptr])
          inuse++
          'debug(string("HeapPtr in stack"),sp)
          quit

    ' search heap
    ifnot inuse
      repeat hp from heaphead to heaptop step 2
        if heapptr == swap16(word[hp])
          'pheap(hp)
          inuse++
          'debug(string("HeapPtr in heap"),hp)
          quit
    
    ifnot inuse
      setFree(nxtptr)
      glen++
'      debug(string("HeapPtr setfree"),heapptr)
'    else
'      debug(string("HeapPtr Found"),heapptr)

    nxtptr := getNext(nxtptr)
    nlen++

'  debug(string("Heap Node Count   "),nlen)
'  debug(string("Garbage Collected "),glen)

'}


