package stamp.core;
import stamp.math.*;

/**
 * This class extends the Timer16 class by scaling the GlobalTimer. Scaling is done
 * by shifting its bits to the right. The number of bits to shift is determined from
 * the selected timeUnit and it provides the largest timer range for that timeUnit.
 *
 * The ScaledTimer16 range is derived from the 32bit global timer.
 * The timer count value is always exact due to the simple shift.
 *
 *                    ScaledTimer16   shift
 *                  <---------------><----->
 *     +--------+--------+--------+--------+
 *     |b31-b24 |b23-b16 | b15-b8 | b7-b0  |
 *     +--------+--------+--------+--------+
 *
 * A ScaledTimer16 would typically be used in code like the following:
 * <pre>
 * ScaledTimer16 t = new ScaledTimer16(400,100); //timeout=400 timeUnits, timeUnit=1msec
 *
 * t.mark(100);
 *
 * while (true) {
 *   // Perform some action.
 *
 *   if ( t.timeout() ) { // Check for a 100 timeUnit timeout.
 *     t.mark(); //uses the stored timeout value
 *     // Perform a periodic action.
 *   }
 * }
 *
 * History
 * V1.1 July 11, 2004
 * Had to move timerScaleMin initialization to constructor due to nullpointer exeception.
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 2.0 December 29, 2006
 */
public class ScaledTimer16 extends Timer16 {

  private int timerShift;     //number of bits to shift CPU timer bits to the right
  private int timeUnit;       //timeUnit in 10 usec units
  private int F;              //used to convert ticks to time
  private int I;              //used to convert time to ticks
  private int K;              //used to convert time to ticks
  private int R;              //remainder from udivide_3216()

  private static int[] timerScaleMin = null; //minimal timeUnit values for scaling

  /**
   * Create a ScaledTimer16 object.
   *
   * @param timeout Number of timeUnits before timeout occurs.
   * @param timeUnit Defines time unit in 10usec units (unsigned).
   *                 A value of 0 is treated as 1.
   *                 The timer tick time is set as in the following table
   * timerShift  Timer16 tick time    timeUnit        recommended timeUnit
   *     0           8.68 usec         1 - 1          1             10 usec
   *     1          17.36              2 - 3          2             20 usec
   *     2          34.72              4 - 6          5             50 usec
   *     3          69.44              7 - 13         10            100 usec
   *     4         138.88             14 - 27         20, 25        200 usec, 250 usec
   *     5         277.76             28 - 55         50            500 usec
   *     6         555.52             56 - 111        100           1 msec
   *     7        1111.04            112 - 222        125, 200      1.25 msec, 2 msec
   *     8        2222.08            223 - 444        250           2.5 msec
   *     9        4444.16            445 - 888        500           5 msec
   *    10        8888.32            889 - 1777       1000, 1250    10 msec, 12.5 msec
   *    11       17776.64           1778 - 3555       2000, 2500    20 msec, 25 msec
   *    12       35553.28           3556 - 7110       5000          50 msec
   *    13       71106.56           7111 - 14221      10000, 12500  0.1 sec, 0.125 sec
   *    14      142213.12          14222 - 28442      20000, 25000  0.2 sec, 0.25 sec
   *    15      284426.24          28443 - 56885      50000         0.5 sec
   *    16      568852.48          56886 - 65535      60000         0.6 sec (0.01 min)
   */
  public ScaledTimer16(int timeout, int timeUnit) {
    super();
    int hi, lo, fr;
    if (timerScaleMin == null) { //only one array is created for all instances
      timerScaleMin = new int[]{1,2,4,7,14,28,56,112,223,445,889,1778,3556,7111,14222,28443,(short)56886};
    }
    if (timeUnit == 0) timeUnit++; //timeUnit must not be 0
    this.timeUnit = timeUnit;      //represents timeUnit * 10 usec
    //find timerShift for the given timeUnit
    if (timeUnit < 0) {
      timerShift = (timeUnit < timerScaleMin[16]) ? 15 : 16;
    }
    else {
      timerShift = 1;
      while (timerShift < 16) {
        if (timeUnit < timerScaleMin[timerShift])
          break;
        timerShift++;
      }
      timerShift--;
    }
    //calculate conversion parameters
    hi = 0;
    lo = (short)56885;
    fr = 16253;
    hi = lo >>> (16-timerShift);
    lo = (lo << timerShift) | (fr >>> (16-timerShift));
    F = udivide_3216(hi,lo,timeUnit); //calculate F
    //System.out.println("R  " + R);
    //System.out.println("F  " + F);
    I = udivide_3216(1,0,F);          //calculate I
    //System.out.println("R  " + R);
    //System.out.println("I  " + I);
    K = udivide_3216(R,0,F);          //calculate K
    //System.out.println("R  " + R);
    //System.out.println("K  " + K);
    mark(timeout); //copy timeout value to superclass
  }

  /**
   * Get current tick count (GlobalTimer shifted right by timerShift bits).
   *
   * @return Current timer tick count (unsigned).
   */
  public int currentTick() {
    int lo = t.tickLo();
    if (timerShift == 0) return lo;
    int hi = t.tickHi();
    if (timerShift == 16) return hi;
    //now rightshift hi:lo by timerShift bits
    return (lo >>> timerShift) | (hi << (16-timerShift));
  }

  /**
   * Convert timer ticks to time units.
   *
   * @param ticks Number of timer ticks (unsigned, as returned by passedTicks() or currentTick()).
   *              ticks range is 0-65535.
   * @return time in timeUnit * 10usec units (unsigned)
   */
  public int TicksToTime(int ticks) {
    //remember F<0
    return UnsignedIntMath.umulf(ticks,F);
  }

  /**
   * Convert time units to timer ticks.
   *
   * @param time Time in timeUnit * 10usec units (unsigned, as returned by TicksToTime()).
   *             time range is 0-TicksToTime(65535).
   * @return Number of timer ticks that represent the time value. (unsigned)
   */
  public int TimeToTicks(int time) {
    //remember I=1 K!=0 or I=2 K=0
    if (I == 2) return time<<1;
    else return time + UnsignedIntMath.umulf(time,K);
  }

  /*
    Ticks to time.
    ==============

           (1 << timerShift) * 8.68           Timer16 tick time             F
    time = ------------------------ * ticks = ----------------- * ticks = ----- * ticks
                 timeUnit * 10                  timeUnit time             65536

                (1 << timerShift) * 8.68   65536 * 8.68   (1 << timerShift)
    F = 65536 * ------------------------ = ------------ * -----------------
                     timeUnit * 10              10            timeUnit

                    (1 << timerShift)                  16253
    F = 56885.248 * -----------------          0.248 = -----
                        timeUnit                       65536

    using registers hi:lo:fr

    hi = 0;
    lo = 56885;
    fr = 16253;

    now shift left by timerShift to multiply with 1<<timerShift

    hi = lo >>> 16-timerShift;
    lo = (lo << timerShift) | (fr >>> 16-timerShift);
    fr = fr << timerShift;

    now divide hi:lo:fr by timeUnit
    fr has no effect on quotient so

          hi:lo
    F = --------       calculation shows:    0x8000 <= F <= 0xFFFF
        timeUnit

    Time to ticks.
    ==============

            65536               R * (65536/F)                  K
    ticks = ----- * time = (I + -------------) * time = (I + -----) * time
              F                 F * (65536/F)                65536

    0x8000 <= F <= 0xFFFF  -->  I=1 and R!=0  --> K!=0
                            or  I=2 and R=0   --> K=0
  */

  /**
   * Divide unsigned 32bit value by unsigned 16bit value.
   * This method was imported from the Int32 class and optimized
   * for this class (quotient 16bit and remainder 16bit).
   *
   * @param nom_hi High word of nominator
   * @param nom_lo Low word of nominator
   * @param denom Denominator
   * @return quotient nom_hi:nom_lo / denom (unsigned)
   *         remainder loaded in global R (unsigned)
   */
  public int udivide_3216(int nom_hi, int nom_lo, int denom) {
    int remHigh=0, quotHigh=0, quotLow=0;
    //clear remainder
    R = 0;
    //left justify the dividend
    int bitCount = 32;
    if (nom_hi == 0) {
      nom_hi = nom_lo;
      bitCount = 16;
    }
    while (nom_hi > 0) {
      nom_hi = (nom_hi << 1) + (nom_lo >>> 15);
      nom_lo = nom_lo << 1;
      bitCount--;
    }
    //process remaining bits in dividend
    for (int i=0; i<bitCount; i++) {
      //rotate dividend left
      int carry = nom_hi >>> 15;
      nom_hi = (nom_hi << 1) + (nom_lo >>> 15);
      nom_lo = nom_lo << 1;
      //rotate remainder left (with carry from dividend)
      remHigh = (remHigh << 1) + (R >>> 15);
      R = (R << 1) + carry;
      //subtract divisor from remainder
      int tmpHigh = remHigh;
      int tmpLow = R - denom;
      if (!CPU.carry())
        tmpHigh--;
      //if divisor < remainder update remainder
      //rotate bit into quotient
      quotHigh = (quotHigh << 1) + (quotLow >>> 15);
      if (tmpHigh >= 0) {
        remHigh = tmpHigh;
        R = tmpLow;
        quotLow = (quotLow << 1) + 1;
      }
      else {
        quotLow = (quotLow << 1);
      }
    }
    return quotLow;
  }

}
