package stamp.core;

/**
 * This class adds stopwatch properties to the ScaledTimer16 class.
 * It provides play, pause, playedTicks and mark methods.
 * Its main purpose is to accumulate time, like a chess clock.
 * Note that the created timer keeps running, even while the stopwatch pauses.
 * So the timeout() method still operates in real time, not in played time.
 *
 * History
 * v1.1 Bugfix
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 July 10, 2004
 */
public class StopwatchScaledTimer16 extends ScaledTimer16 {

  private int playedTicks;       //number of ticks played since last call to mark()
  private int pausedTicks;       //number of ticks paused since last call to mark()
  private boolean play;          //play (true) or pause (false)

  /**
   * Create a StopwatchScaledTimer16 object.
   * The stopwatch starts paused.
   *
   * @param timeout Number of timeUnits before timeout occurs.
   * @param timeUnit Defines time unit in 10usec units (unsigned).
   *                 A value of 0 is treated as 1.
   *                 The timer tick time is set as in the following table
   * timerShift  Timer16 tick time    timeUnit        recommended timeUnit
   *     0           8.68 usec         1 - 1          1             10 usec
   *     1          17.36              2 - 3          2             20 usec
   *     2          34.72              4 - 6          5             50 usec
   *     3          69.44              7 - 13         10            100 usec
   *     4         138.88             14 - 27         20, 25        200 usec, 250 usec
   *     5         277.76             28 - 55         50            500 usec
   *     6         555.52             56 - 111        100           1 msec
   *     7        1111.04            112 - 222        125, 200      1.25 msec, 2 msec
   *     8        2222.08            223 - 444        250           2.5 msec
   *     9        4444.16            445 - 888        500           5 msec
   *    10        8888.32            889 - 1777       1000, 1250    10 msec, 12.5 msec
   *    11       17776.64           1778 - 3555       2000, 2500    20 msec, 25 msec
   *    12       35553.28           3556 - 7110       5000          50 msec
   *    13       71106.56           7111 - 14221      10000, 12500  0.1 sec, 0.125 sec
   *    14      142213.12          14222 - 28442      20000, 25000  0.2 sec, 0.25 sec
   *    15      284426.24          28443 - 56885      50000         0.5 sec
   *    16      568852.48          56886 - 65535      60000         0.6 sec (0.01 min)
   */
  public StopwatchScaledTimer16(int timeout, int timeUnit) {
    super(timeout,timeUnit);
    this.mark();
  }

  /**
   * Play stopwatch.
   */
  public void play() {
    if (!play) { //skip if already playing
      pausedTicks = (passedTicks() - playedTicks);
      play = true;
    }
  }

  /**
   * Pause stopwatch.
   */
  public void pause() {
    if (play) { //skip if already paused
      playedTicks = (passedTicks() - pausedTicks);
      play = false;
    }
  }

  /**
   * Get number of stopwatch ticks played.
   *
   * @return Number of stopwatch ticks played since last call to mark();
   *         Use TicksToTime() to convert the returned value to timeUnits.
   *
   */
  public int playedTicks() {
    return (!play) ? playedTicks : (passedTicks() - pausedTicks);
  }

  /**
   * Mark stopwatch.
   * This pauses the stopwatch and clears the number of ticks played.
   * It marks the current timer count.
   *
   * @param timeout Number of timeUnits before timeout occurs.
   */
  public void mark(int timeout) {
    super.mark(timeout);
    this.mark();
  }

  /**
   * Mark stopwatch.
   * This pauses the stopwatch and clears the number of ticks played.
   * It marks the current timer count. Uses the last stored timeout value.
   */
  public void mark() {
    playedTicks = 0;
    pausedTicks = 0;
    play = false;
    super.mark();
  }

}