package stamp.core;

/**
 * This class provides a 16bit timer derived from the global 32bit timer.
 * The calculation of the stop time has been moved from the timeout() method to
 * the mark() method to increase speed. Mark() is usually called once, while timeout()
 * is called frequently to test for the timeout.
 * This class is to be extended. Extended classes must provide 3 methods:
 *   currentTick() that returns the current timer count
 *   TicksToTime() that converts timer ticks into timeUnits
 *   TimeToTicks() that converts timeUnits into timer ticks
 *
 * Timeout areas.
 * The following shows the different possibilities on start and stop being < 0 or > 0.
 * S marks start and P marks stop. Timeout time equals stop - start. A timeout occurs
 * whenever the current time is in a marked area.
 *
 *          S    P
 *     +----------------+----------------+      timeout if
 *     |////|    |//////|////////////////|      (curr >= stop) || (curr < start)
 *     +----------------+----------------+
 *     0             0x8000           0xFFFF
 *
 *          P    S
 *     +----------------+----------------+      timeout if
 *     |    |////|      |                |      (curr >= stop) && (curr < start)
 *     +----------------+----------------+
 *     0             0x8000           0xFFFF
 *
 *                             S    P
 *     +----------------+----------------+      timeout if
 *     |////////////////|//////|    |////|      (curr >= stop) || (curr < start)
 *     +----------------+----------------+
 *     0             0x8000           0xFFFF
 *
 *                             P    S
 *     +----------------+----------------+      timeout if
 *     |                |      |////|    |      (curr >= stop) && (curr < start)
 *     +----------------+----------------+
 *     0             0x8000           0xFFFF
 *
 *          S                  P
 *     +----------------+----------------+      timeout if
 *     |////|           |      |/////////|      (curr >= stop) && (curr < start)
 *     +----------------+----------------+
 *     0             0x8000           0xFFFF
 *
 *          P                  S
 *     +----------------+----------------+      timeout if
 *     |    |///////////|//////|         |      (curr >= stop) || (curr < start)
 *     +----------------+----------------+
 *     0             0x8000           0xFFFF
 *
 * Analyzing the above yields the following timeout test:
 *
 *     if (start < stop) timeout = (curr >= stop) || (curr < start);
 *     else timeout = (curr >= stop) && (curr < start);
 *
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 2.0 December 29, 2006
 */
public abstract class Timer16 {

  public abstract int currentTick();
  public abstract int TimeToTicks(int time);
  public abstract int TicksToTime(int ticks);

  private int start;         //marked current time
  private int timeout_ticks; //marked timeout value
  private int stop;          //equals start+timeout
  protected static Timer t;  //to access tickLo() and tickHi()

  /**
   * Create a Timer16 object.
   */
  public Timer16() {
    if (t == null) t = new Timer();
  }

  /**
   * Remember the current time and timeout value
   * and calculate the stop time.
   *
   * @param timeout Number of timeUnits before timeout occurs.
   */
  public void mark(int timeout) {
    timeout_ticks = TimeToTicks(timeout);
    mark();
  }

  /**
   * Remember the current time and calculate the stop time.
   * Uses the last stored timeout value.
   */
  public void mark() {
    start = currentTick();
    stop = start + timeout_ticks;
  }

  /**
   * Get the number of passed ticks since the last call to mark().
   *
   * @return Passed time in ticks (unsigned)
   */
  public int passedTicks() {
    return currentTick() - start;
  }

  /**
   * Get the passed time in timeUnits since the last call to mark().
   *
   * @return Passed time in timeUnits (unsigned).
   */
  public int passedTime() {
    return TicksToTime(passedTicks());
  }

  /**
   * Test for timeout using the start and stop time remembered
   * by the last call to mark().
   *
   * @return true if the timeout period has expired.
   */
  public boolean timeout() {
    int curr = currentTick();
    return (start < stop) ? ((curr >= stop) || (curr < start)) : ((curr >= stop) && (curr < start));
  }

}