/* Adapted version of Timer.java
 * June 16, 2004: Added passedMS() method
 *                Author: Peter Verkaik (peterverkaik@boselectro.nl)
 */

/*
 * Copyright  1999-2002 Celsius Research Ltd. All rights reserved.
 * Do not modify this file.
 */

package stamp.core;

/**
 * A general purpose 32 bit timer.
 *
 * There
 * is only a single instantiation of the timer itself, but there can be
 * multiple <code>Timer</code> objects which use the timer value.
 * The first <code>Timer</code> created starts the timer VP. Subsequent
 * instantiations use the same VP. Caution should be taken that stopping
 * the timer VP does not affect other <code>Timer</code> objects.
 * <p>
 * The timer is a 32 bit counter which is incremented every 8.68us. Therefore:
 * <ul>
 *    <li>1 millisecond = 115</li>
 *    <li>1 second = 115,207</li>
 *    <li>1 minute = 6,912,442</li>
 *    <li>1 hour = 414,746,543</li>
 * </ul>
 * The timer will overflow every 10.36 hours.
 * <p>
 * A timer would typically be used in code like the following:
 * <pre>
 * Timer t = new Timer();
 *
 * t.mark();
 *
 * while (true) {
 *   // Perform some action.
 *
 *   if ( t.timeout(100) ) { // Check for a 100ms timeout.
 *     t.mark();
 *     // Perform a periodic action.
 *   }
 * }
 *
 * </pre>
 * @author Chris Waters
 * @version 1.0 1/25/00
 */
public class Timer extends VirtualPeripheral {

  /**
   * Creates a <code>Timer</code> object. The first instantiation of this class
   * starts the timer VP. Subsequent instantiations use the same VP.
   */
  public Timer() {
    if ( globalTimerVP == null )
      start(this);
  }

  /**
   * Starts the timer VP. The constructor will start the Timer when it is first
   * created. If it is later stopped with <code>stop()</code> then
   * <code>start()</code> will start it again with the same set of parameters.
   * <p>
   * It is an error to call start if the Timer is already running.
   */
  public static void start(Timer vp) {
    // Install the interrupt handler.
    CPU.installVP( vp );
    globalTimerVP = vp;
    globalVPBank = vp.vpBank;

    // Clear the timer.
    CPU.writeRegister( globalVPBank|nmTimer1, 0 );
    CPU.writeRegister( globalVPBank|nmTimer2, 0 );
    CPU.writeRegister( globalVPBank|nmTimer3, 0 );
    CPU.writeRegister( globalVPBank|nmTimer4, 0 );

    // Start the timer.
    init(globalVPBank);
  }

  /**
   * Stops the timer VP.
   */
  public static void stop() {
    if ( globalTimerVP != null )
      CPU.removeVP(globalTimerVP);
    globalTimerVP = null;
  }

  /**
   * Retrieves the high 16 bits of the timer value.
   *
   * @return the high 16 bits of the timer value.
   */
  public int tickHi() {
    return (CPU.readRegister( globalVPBank|nmLatchTimer4 )<<8)|CPU.readRegister( globalVPBank|nmLatchTimer3 );
  }

  /**
   * Retrieves the low 16 bits of the timer value.
   * <p>
   * <code>tickLo()</code> should always be called before <code>tickHi()</code>.
   *
   * @return the low 16 bits of the timer value.
   */
  public int tickLo() {
    Timer.latch(globalVPBank);
    return (CPU.readRegister( globalVPBank|nmLatchTimer2 )<<8)|CPU.readRegister( globalVPBank|nmLatchTimer1 );
  }

  /**
   * Remembers the current timer value. Call this method to mark the start of
   * a timeout period.
   */
  public void mark() {
    startLo = tickLo();
    startHi = tickHi();
  }

  /**
   * Calculate the time passed since the last call to the <code>mark()</code> method.
   * The returned value is in millisecond units.
   *
   * @return Passed time in milliseconds (unsigned)
   * </code by Peter Verkaik (peterverkaik@boselectro.nl)>
   */
  public int passedMS() {
    int l = tickLo(); //current time (unsigned)
    int h = tickHi();
    int r;
    h = h - startHi; //calculate passed time = current time - mark time
    if (startLo != 0) {
      l = l - startLo;
      if (!CPU.carry()) h--;
    }
    // now convert h:l into milliseconds: timeMS = 569*h + (l/115)
    r = (569*h)+(l/115);
    if (l<0) r += 569;   //l is really unsigned, so if l<0 (0x8000-0xFFFF) l/115 equals ((65536+l)/115) = 569 + l/115
    return r;
  }

  /**
   * Compares the current timer value with the value remembered by the last
   * call to the <code>mark()</code> method. If the difference is greater than
   * the parameter then returns true.
   *
   * @return true if the timeout period has been exceeded.
   */
  public boolean timeout(int hi, int lo) {
    int l = tickLo(); //current time (unsigned)
    int h = tickHi();

    int ah = startHi; //calculate timeout time = mark time + timeout value
    int al = startLo + lo;
    if (CPU.carry()) ah++;
    ah += hi;

    // now ah:al holds the timeout time (unsigned)
    // we must return true if current time >= timeout time

    if (ah > h) {
      return false;
    }
    else {
      if (ah == h) {
        if ( ( al < 0 && l < 0) || ( al >= 0 && l >= 0 )) {
          return (al <= l);
        }
        else {
          if ( al < 0 && l > 0 ) return false;
        }
      }
    }
    return true;
  }

  /**
   * Checks whether <code>timeMS</code> milliseconds have elapsed since
   * the last call to <code>mark()</code>.
   *
   * @return true if the timeout period has been exceeded.
   */
  public boolean timeout(int timeMS) {
    return timeout(timeMS/569, (timeMS%569)*115);
  }

  /**
   * Checks whether <code>timeS</code> seconds have elapsed since
   * the last call to <code>mark()</code>. The maximum timeout is 935 seconds.
   *
   * @return true if the timeout period has been exceeded.
   */
  public boolean timeoutSec(int timeS) {
    return timeout((timeS*18)/10, 0);
  }

//============================================================================
// Private methods and fields below this point.
//============================================================================

  protected static Timer globalTimerVP = null;
  protected static int globalVPBank;
  int startHi, startLo;

  protected static native void init(int bank);
  protected static native void latch(int bank);

  final static int nmPeriphISRVec = 0;

  final static int nmLatchTimer1 = 6;
  final static int nmLatchTimer2 = 7;
  final static int nmLatchTimer3 = 8;
  final static int nmLatchTimer4 = 9;

  final static int nmTimer1 = 2;
  final static int nmTimer2 = 3;
  final static int nmTimer3 = 4;
  final static int nmTimer4 = 5;
}