/* This is an adapted version of the Uart class that adds support for
 * clearing and accessing the buffer. Speeded up buffer access by declaring
 * buffer as a byte array, rather than an int array. Made the buffer public
 * for even faster access. Provided a bufferSet method to fill buffer
 * without immediately starting transmission. Added parity support.
 * Date: 10-22-2004, Peter Verkaik (peterverkaik@boselectro.nl)
 */

/*
 * Copyright  1999-2002 Celsius Research Ltd. All rights reserved.
 * Do not modify this file.
 */

package stamp.core;

/**
 * A virtual peripheral UART. Each instance of this class provides
 * asynchronous serial communication in a single direction. For full duplex
 * communications create two <code>Uart</code> objects, one for each direction.
 * <p>
 * The Uart class uses 256 byte buffers for both transmit and receive. The buffer
 * size is not configurable. If a handshaking pin is specified for a receive Uart then
 * the peer will be signalled to stop transmitting when there are 16 bytes still free
 * in the receive buffer. This handshaking is done totally by the uart VP and cannot
 * be controlled.
 *
 * @author Chris Waters
 * @author Nick Kelsey
 * Date 04-28-2003
 */

public class Uart extends VirtualPeripheral {

  /**
   * The value to be using in the direction parameter to create a receiver.
   * The receiver starts automatically.
   */
  public final static int dirReceive = 1;

  /**
   * The value to be using in the direction parameter to create a transmitter.
   * The transmitter starts automatically.
   */
  public final static int dirTransmit = 0;

  /**
   * The value to be using in the direction parameter to create a receiver.
   * The receiver is stopped.
   */
  public final static int dirReceiveStopped = 3;

  /**
   * The value to be using in the direction parameter to create a transmitter.
   * The transmitter is stopped.
   */
  public final static int dirTransmitStopped = 2;

  /*
   * The baud rate speed values are computed by:
   *
   *   speed = 25,000,000/(baud rate * 217)
   */

  /**
   * The speed value for 600 baud.
   */
  public final static int speed600 = 192;

  /**
   * The speed value for 1200 baud.
   */
  public final static int speed1200 = 96;

  /**
   * The speed value for 2400 baud.
   */
  public final static int speed2400 = 48;

  /**
   * The speed value for 4800 baud.
   */
  public final static int speed4800 = 24;

  /**
   * The speed value for 7200 baud.
   */
  public final static int speed7200 = 16;

  /**
   * The speed value for 9600 baud.
   */
  public final static int speed9600 = 12;

  /**
   * The speed value for 14400 baud.
   */
  public final static int speed14400 = 8;

  /**
   * The speed value for 19200 baud.
   */
  public final static int speed19200 = 6;

  /**
   * The speed value for 28800 baud.
   */
  public final static int speed28800 = 4;

  /**
   * The speed value for 38400 baud.
   */
  public final static int speed38400 = 3;

  /**
   * The speed value for 57600 baud.
   */
  public final static int speed57600 = 2;

  /**
   * The speed value for 115200 baud.
   */
  public final static int speed115200 = 1;

  /**
   * The value for dataInvert and hsInvert to indicate that inverted logic
   * should be used.
   */
  public final static boolean invert = true;

  /**
   * The value for dataInvert and hsInvert to indicate that inverted logic
   * should not be used.
   */
  public final static boolean dontInvert = false;

  /**
   * Value to generate 1 stop bit.
   */
  public final static int stop1 = 1;

  /**
   * Value to generate 2 stop bits.
   */
  public final static int stop2 = 3;

  /**
   * Value to generate 3 stop bits.
   */
  public final static int stop3 = 7;

  /**
   * Value to generate 4 stop bits.
   */
  public final static int stop4 = 15;

  /**
   * Value to generate 5 stop bits.
   */
  public final static int stop5 = 31;

  /**
   * Value to generate 6 stop bits.
   */
  public final static int stop6 = 63;

  /**
   * Value to generate 7 stop bits.
   */
  public final static int stop7 = 127;

  /**
   * Value to generate 8 stop bits.
   */
  public final static int stop8 = 255;

  /**
   * Creates and initialises a new Uart for simplex communication
   * using no handshaking. The constructor calls the <code>start()</code>
   * method so the Uart will begin working immediately.
   *
   * @param direction  whether the Uart is to be used in transmit or receive mode.
   *                   Valid options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public Uart(int direction, int dataPin, boolean dataInvert, int baudRate, int stopBits) {
    this(direction,UartTools.NONE,dataPin,dataInvert,0,dontInvert,baudRate,stopBits);
  }

  /**
   * Creates and initialises a new Uart for simplex communication
   * using no handshaking. The constructor calls the <code>start()</code>
   * method so the Uart will begin working immediately.
   * This constructor lets you specify parity.
   *
   * @param direction  whether the Uart is to be used in transmit or receive mode.
   *                   Valid options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param parity     UartTools.NONE, UartTools.ODD or UartTools.EVEN
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public Uart(int direction, int parity, int dataPin, boolean dataInvert, int baudRate, int stopBits) {
    this(direction,parity,dataPin,dataInvert,0,dontInvert,baudRate,stopBits);
  }

  /**
   * Creates and initialises a new Uart for simplex communication using
   * hardware (RTS/CTS) handshaking. The constructor
   * calls the <code>start()</code> method so the Uart will begin working immediately.
   *
   * @param direction  whether the Uart is to be used transmit or receive mode. Valid
   *                   options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param hsPin      the pin to use for hardware handshaking.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public Uart(int direction, int dataPin, boolean dataInvert, int hsPin, int baudRate, int stopBits) {
    this(direction,UartTools.NONE,dataPin,dataInvert,hsPin,dataInvert,baudRate,stopBits);
  }

  /**
   * Creates and initialises a new Uart for simplex communication
   * using hardware (RTS/CTS) handshaking. The constructor calls the <code>start()</code>
   * method so the Uart will begin working immediately.
   *
   * @param direction  whether the Uart is to be used transmit or receive mode. Valid
   *                   options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param hsPin      the pin to use for hardware handshaking.
   * @param hsInvert   whether the logic of the handshaking pin should be inverted or
   *                   not.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public Uart(int direction, int dataPin, boolean dataInvert, int hsPin, boolean hsInvert, int baudRate, int stopBits) {
    this(direction,UartTools.NONE,dataPin,dataInvert,hsPin,hsInvert,baudRate,stopBits);
  }

  /**
   * Creates and initialises a new Uart for simplex communication
   * using hardware (RTS/CTS) handshaking. The constructor calls the <code>start()</code>
   * method so the Uart will begin working immediately.
   * This constructor lets you specify parity.
   *
   * @param direction  whether the Uart is to be used transmit or receive mode. Valid
   *                   options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param parity     UartTools.NONE, UartTools.ODD or UartTools.EVEN
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param hsPin      the pin to use for hardware handshaking.
   * @param hsInvert   whether the logic of the handshaking pin should be inverted or
   *                   not.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public Uart(int direction, int parity, int dataPin, boolean dataInvert, int hsPin, boolean hsInvert, int baudRate, int stopBits) {
    this.direction = direction&1; //only remember dirTransmit or dirReceive
    this.parity = parity;
    this.dataPin = dataPin;
    this.dataInvert = dataInvert;
    this.hsPin = hsPin;
    this.hsInvert = hsInvert;
    this.baudRate = baudRate;
    this.stopBits = stopBits;
    // Allocate the buffer.
    buffer = new char[stdBufferSize];
    start(); //make sure uart is initialized
    if ((direction&2)!=0) stop(); //only start if dirReceive and dirTransmit
  }

  /**
   * Starts the Uart running. If it is later stopped with <code>stop()</code>
   * then <code>start()</code> will start it again with the same set of parameters.
   */
  public void start() {
    if (started) return; //already started
    //install the interrupt handler.
    CPU.installVP(this);
    //configure the VP registers.
    CPU.writeRegister(vpBank|nmUartPort, dataPin>>8);
    CPU.writeRegister(vpBank|nmUartPin, dataPin);
    CPU.writeRegister(vpBank|nmUartHSPort, hsPin>>8);
    CPU.writeRegister(vpBank|nmUartHSPin, hsPin);
    CPU.writeRegister(vpBank|nmUartInvert, (dataInvert?1:0)|(hsInvert?2:0));
    CPU.writeRegister(vpBank|nmUartBaud, baudRate);
    CPU.writeRegister(vpBank|nmUartStopBits, stopBits);
    CPU.writeObject(vpBank|nmUartBufferH, buffer);
    //initialise the Uart VP.
    if (direction == dirTransmit) {
      txInit(vpBank);
    }
    else {
      rxInit(vpBank);
    }
    //clear the buffer
    bufferClear();
    started = true;
  }

  /**
   * Stops the Uart.
   */
  public void stop() {
    if (!started) return; //already stopped
    while (!sendBufferEmpty())
      ;  //wait until character transmitted
    CPU.removeVP(this);
    started = false;
  }

  /**
   * Change transmission direction.
   *
   * @param direction new direction
   */
  public void setDirection(int direction) {
    boolean started = this.started;
    stop();
    this.direction = direction&1;
    if (started) start(); //only start if already started
  }

  /**
   * Reinitialises a Uart for simplex communication using
   * no handshaking. The Uart will begin working immediately.
   *
   * @param direction  whether the Uart is to be used transmit or receive mode. Valid
   *                   options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public void restart(int direction, int dataPin, boolean dataInvert, int baudRate, int stopBits) {
    restart(direction,UartTools.NONE,dataPin,dataInvert,0,dontInvert,baudRate,stopBits);
  }

  /**
   * Reinitialises a Uart for simplex communication using
   * no handshaking. The Uart will begin working immediately.
   * This constructor lets you specify parity.
   *
   * @param direction  whether the Uart is to be used transmit or receive mode. Valid
   *                   options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param parity     UartTools.NONE, UartTools.ODD or UartTools.EVEN
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public void restart(int direction, int parity, int dataPin, boolean dataInvert, int baudRate, int stopBits) {
    restart(direction,parity,dataPin,dataInvert,0,dontInvert,baudRate,stopBits);
  }

  /**
   * Reinitialises the Uart for simplex communication using
   * hardware (RTS/CTS) handshaking. The Uart will begin working immediately.
   *
   * @param direction  whether the Uart is to be used transmit or receive mode. Valid
   *                   options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param hsPin      the pin to use for hardware handshaking.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public void restart(int direction, int dataPin, boolean dataInvert, int hsPin, int baudRate, int stopBits) {
    restart(direction,UartTools.NONE,dataPin,dataInvert,hsPin,dataInvert,baudRate,stopBits);
  }

  /**
   * Reinitialises the Uart for simplex communication using
   * hardware (RTS/CTS) handshaking. The Uart will begin working immediately.
   *
   * @param direction  whether the Uart is to be used transmit or receive mode. Valid
   *                   options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param hsPin      the pin to use for hardware handshaking.
   * @param hsInvert   whether the logic of the handshaking pin should be inverted or
   *                   not.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public void restart(int direction, int dataPin, boolean dataInvert, int hsPin, boolean hsInvert, int baudRate, int stopBits) {
    restart(direction,UartTools.NONE,dataPin,dataInvert,hsPin,hsInvert,baudRate,stopBits);
  }

  /**
   * Reinitialises a Uart for simplex communication using
   * hardware (RTS/CTS) handshaking. The Uart will begin working immediately.
   * This constructor lets you specify parity.
   *
   * @param direction  whether the Uart is to be used transmit or receive mode. Valid
   *                   options are <code>Uart.dirTransmit</code> and
   *                   <code>Uart.dirReceive</code>.
   * @param parity     UartTools.NONE, UartTools.ODD or UartTools.EVEN
   * @param dataPin    the Javelin Stamp I/O pin to use for sending or receiving data.
   * @param dataInvert whether the data should be inverted or not. Valid options are
   *                   <code>Uart.invert</code> and <code>Uart.dontInvert</code>.
   * @param hsPin      the pin to use for hardware handshaking
   * @param hsInvert   whether the logic of the handshaking pin should be inverted or
   *                   not.
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public void restart(int direction, int parity, int dataPin, boolean dataInvert, int hsPin, boolean hsInvert, int baudRate, int stopBits) {
    stop();
    this.direction = direction&1;
    this.parity = parity;
    this.dataPin = dataPin;
    this.dataInvert = dataInvert;
    this.hsPin = hsPin;
    this.hsInvert = hsInvert;
    this.baudRate = baudRate;
    this.stopBits = stopBits;
    if ((direction&2)==0) start(); //only start if dirReceive and dirTransmit
  }

  /**
   * Change the non-physical parameters.
   *
   * @param parity     UartTools.NONE, UartTools.ODD or UartTools.EVEN
   * @param baudRate   the baud rate to use.
   * @param stopBits   the number of stop bits to use. Valid values are between
   *                   <code>Uart.stop1</code> and <code>Uart.stop8</code>.
   */
  public void setMode(int parity, int baudRate, int stopBits) {
    boolean started = this.started;
    stop();
    this.parity = parity;
    this.baudRate = baudRate;
    this.stopBits = stopBits;
    if (started) start(); //only start if already started
  }

  /**
   * Check space in the transmit buffer.
   *
   * @return transmit buffer is full
   */
  public boolean sendBufferFull() {
    return (bufferCount() >= 254);
  }

  /**
   * Check space in the transmit buffer.
   *
   * @return transmit buffer is empty (and nothing being sent now)
   */
  public boolean sendBufferEmpty() {
    return (CPU.readRegister(vpBank|nmUartDataL) == 0)
        && (CPU.readRegister(vpBank|nmUartDataH) == 0)
        && (bufferCount() == 0);
  }

  /**
   * Check buffer use
   *
   * @return number of bytes in buffer.
   */
  public int bufferCount() {
    return (CPU.readRegister(vpBank|nmUartHead)
            - CPU.readRegister(vpBank|nmUartTail))&0xff;
  }

  /**
   * Clear buffer.
   */
  public void bufferClear() {
    int headPointer = CPU.readRegister(vpBank|nmUartHead);
    //make tailPointer == headPointer
    CPU.writeRegister(vpBank|nmUartTail,headPointer);
  }

  /**
   * Set the number of bytes in the buffer.
   * This method is intended to set the number of characters
   * in the buffer AFTER the buffer has been filled.
   * Before filling the buffer it should be empty.
   *
   * @param bytes The number of bytes to transmit.
   *        The number of bytes in the buffer equals head-tail.
   */
  public void bufferSet(int bytes) {
    int headPointer = CPU.readRegister(vpBank|nmUartHead);
    //make headPointer == headPointer + bytes
    //Start filling the buffer at headPointer, incrementing the
    //pointer while filling. After the last byte is written,
    //the pointer equals the new headpointer.
    CPU.writeRegister(vpBank|nmUartHead,headPointer+bytes);
  }

  /**
   * This is the main transmit method.
   * Adds a byte to the transmit buffer. The byte is queued for transmit at the
   * end of the buffer. If the buffer is already full then <code>writeByte()</code>
   * returns -1.
   *
   * @param data The data to transmit.
   *             Only the lower 8 bits of the parameter are used.
   * @return Lower 8 bits of parameter data
   *         -1 if buffer full.
   */
  public int writeByte(int data) {
    int b;
    /* may we send?
     * Note: if we may not send, return immediately. Do not store
     * character in buffer even if there is space. Any bytes already
     * in the buffer will be transmitted by the VP.
    */
    if (sendBufferFull()) return -1;
    int headPointer = CPU.readRegister(vpBank|nmUartHead);
    if (parity != 0) {
      b = (parity == UartTools.ODD) ? UartTools.oddParity(data) : UartTools.evenParity(data);
    }
    else b = data;
    buffer[headPointer++] = (char)b;
    CPU.writeRegister(vpBank|nmUartHead,headPointer);
    return (data & 0xff);
  }

  /**
   * Transmits a String.
   * This method returns -1 if there is not enough room in the buffer
   * to hold the string.
   *
   * @param data the string to transmit.
   * @return The length of the string or -1.
   */
  public int writeString(String data) {
    return writeString(data.toCharArray(),data.length(),0);
  }

  /**
   * Transmits a String.
   * This method returns -1 if there is not enough room in the buffer
   * to hold the string.
   *
   * @param data the string to transmit.
   * @param num Number of bytes to transmit.
   * @param offset Index in data to start from.
   * @return The number of bytes transmitted or -1.
   */
  public int writeString(char[] data, int num, int offset) {
    int i,bytes;
    bytes = bufferCount();
    if (num > (stdBufferSize-bytes-2)) return -1; //string does not fit
    sendString(data,num,offset);
    return num;
  }

  /**
   * Adds a byte to the transmit buffer. The byte is queued for transmit at the
   * end of the buffer. If the buffer is already full then <code>sendByte()</code>
   * will block until there is room in the buffer.
   *
   * @param data The data to transmit.
   *             Only the lower 8 bits of the parameter are used.
   */
  public void sendByte(int data) {
    while (writeByte(data) == -1) ; //wait until data stored in transmit buffer
  }

  /**
   * Transmits a String. This method blocks until the string is
   * completely written to the transmit buffer.
   *
   * @param data the string to transmit.
   */
  public void sendString(String data) {
    sendString(data.toCharArray(),data.length(),0);
  }

  /**
   * Transmits a String. This method blocks until the string is
   * completely written to the transmit buffer.
   *
   * @param data the string to transmit.
   */
  public void sendString(char[] data) {
    int c,i;
    i = 0;
    while ((c = data[i++]) != 0) sendByte(c);
  }

  /**
   * Transmits a String. This method blocks until the string is
   * completely written to the transmit buffer.
   *
   * @param data the character array to transmit.
   * @param num Number of bytes to transmit.
   * @param offset Index in data to start from.
   */
  public void sendString(char[] data, int num, int offset) {
    while (num > 0) {
      sendByte(data[offset++]);
      num--;
    }
  }

  /**
   * This is the main receive method.
   * Receives a byte from a receive Uart. This method returns -1
   * if no byte is available.
   *
   * @return the next byte in the receive buffer.
   *         If parity is used, the byte returns with bit7 set if parity error
   *         -1 if buffer empty.
   */
  public int readByte() {
    int b; boolean error;
    int headPointer = CPU.readRegister(vpBank|nmUartHead);
    int tailPointer = CPU.readRegister(vpBank|nmUartTail);
    if (headPointer == tailPointer) return -1;
    b = buffer[tailPointer++]&0xff;
    CPU.writeRegister(vpBank|nmUartTail,tailPointer);
    // Inform the receiver so it can do the handshaking.
    rxRead(vpBank);
    if (parity != 0) { //check for parity error
      error = (parity == UartTools.ODD) ? !UartTools.isOddParity(b) : !UartTools.isEvenParity(b);
      b &= 0x7F; //normally bit7 is clear when using parity
      if (error) b |= 0x80; //set bit7 to indicate parity error
    }
    return b;
  }

  /**
   * Receive a String.
   *
   * @param data the character array to hold the string.
   * @param num Maximum number of bytes to receive.
   * @param offset Index in data to start from.
   * @return The number of bytes actually placed into data.
   */
  public int readString(char[] data, int num, int offset) {
    int i=0;
    if (num < 1) return 0;
    int c = readByte();
    while ((c != -1) && (num > 0)) {
      data[offset++] = (char)c;
      num--;
      c = readByte();
      i++;
    }
    return i;
  }

  /**
   * Receives a byte from a receive Uart. This method will block until
   * a byte is available.
   *
   * @return the next byte in the receive buffer.
   */
  public int receiveByte() {
    int b;
    while ((b=readByte()) == -1) ;
    return b;
  }

  /**
   * Receive a String. This method blocks until num characters are received.
   *
   * @param data the character array to hold the string.
   * @param num Maximum number of bytes to receive.
   * @param offset Index in data to start from.
   */
  public void receiveString(char[] data, int num, int offset) {
    for (int i=0; i<num; i++) data[offset++] = (char)receiveByte();
  }

  /**
   * Test whether any bytes are available in the receive buffer.
   * Can also be used to test if the transmit buffer is empty.
   *
   * @return true if there is at least one byte available in the receive buffer,
   * or at least one byte free in the transmit buffer.
   */
  public boolean byteAvailable() {
    return (bufferCount() > 0);
  }

//============================================================================
// Private methods and fields below this point.
//============================================================================

  int direction;        //dirTransmit or dirReceive
  int dataPin;          //pin for data transmit and receive
  boolean dataInvert;   //logic level dataPin
  int hsPin;            //handshake pin
  boolean hsInvert;     //logic level hsPin
  int baudRate;         //speed600 - speed57600
  int stopBits;         //stop1 - stop8
  int parity;           //0=none, 1=odd, 2=even
  boolean started;      //true if started

  static native void txInit(int bank);
  static native void rxInit(int bank);
  static native void rxRead(int bank);

  final static int nmUartDataH    = 0x2;
  final static int nmUartDataL    = 0x3;

  final static int nmUartInvert   = 0x5; // 0 for normal, 1 for invert
  final static int nmUartBaud     = 0x6; // 1000000/(bps * 4.34)
  final static int nmUartStopBits = 0x7; // Tx only: 00000001 for 1 stop bit, 00000011 for 2 etc

  final static int nmUartPort     = 0xC; // 0 for Port0, 1 for Port1
  final static int nmUartPin      = 0xD; // bit mask (1 of 8 set)
  final static int nmUartHSPort   = 0xE; // 0 for Port0, 1 for Port1
  final static int nmUartHSPin    = 0xF; // bitmask (1 of 8 set)

  final static int nmUartBufferH  = 0xA;
  final static int nmUartBufferL  = 0xB;

  //Index of the next free byte in the buffer.
  final static int nmUartHead     = 0x8;

  //Index of the oldest byte in the buffer.
  //If tailPointer == headPointer the buffer is empty.
  final static int nmUartTail     = 0x9;

  //Transmit and receive buffer storage.
  final static int stdBufferSize  = 256; //bytes
  public char buffer[]; //Defined public to allow fast access.

}
