package stamp.filesystem;

/**
 * This class defines the methods that the FileSystem class requires to function.
 * These methods perform the hardware I/O.
 *
 * <p><pre>
 * Revision History:
 * September 13, 2004 v1.0: Initial release.
 * </pre><p>
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 September 13, 2004
 */

public abstract class FSinterface {

  public int SECTORSIZE;     //number of bytes per sector (must be power of 2)
  protected int SECTORBITS;  //base 2 power of sectorsize (eg. sectorsize = 1<<bits)

  /**
   * Constructor for FSinterface.
   *
   * @param sectorsize Number of bytes in sector (must be power of 2).
   */
  protected FSinterface(int sectorsize) {
    int i;
    SECTORSIZE = sectorsize;
    for (SECTORBITS=0; SECTORBITS<16; SECTORBITS++) {
      if ((sectorsize&1) != 0) break;
      sectorsize >>>= 1;
    }
  }

  /**
   * Test if filesystem device present.
   *
   * @return True if filesystem device present.
   */
  protected abstract boolean present();

  /**
   * Test if filesystem device ready.
   *
   * @return True if filesystem device ready.
   */
  protected abstract boolean ready();

  /**
   * Reset device.
   * If the device does not support this it must return true.
   *
   * @return True if reset succesful.
   */
  protected abstract boolean reset();

  /**
   * Get filesystem status.
   *
   * @return errorcode.
   *         Returns 0 if filesytem ok.
   */
  protected abstract int status();

  /**
   * Get number of files.
   *
   * @return Number of files.
   */
  protected abstract int files();

  /**
   * Get filename.
   *
   * @param index Logical entry for file.
   *              Range is 0 to files()-1.
   * @return Name of file as ASCIIZ or null.
   */
  protected abstract char[] fileName(int index);

  /**
   * Search a file.
   *
   * @param name Path/name of the file.
   *             See extended class for filename specifications.
   * @return File index or errorcode.
   *         This file index identifies a file on a physical media.
   *         The transferring to/from media is handled by the extended class.
   */
  protected abstract int searchFile(char[] name);

  /**
   * Create a file.
   * The size of the created file is 0.
   * The name is already checked and no file with that name was found.
   *
   * @param name Path/name of the file.
   *             See extended class for filename specifications.
   * @return File index or errorcode.
   *         This file index identifies a file on a physical media.
   *         The transferring to/from media is handled by the extended class.
   */
  protected abstract int createFile(char[] name);

  /**
   * Delete a file.
   * The file index is valid.
   *
   * @param fi File index of file to delete.
   * @return errorcode.
   */
  protected abstract int deleteFile(int fi);

  /**
   * Close a file.
   * The new size will be written to disk
   * The file index is valid.
   *
   * @param fi File index of file to close.
   * @param size_hi High word of new filesize
   * @param size_lo Low word of new filesize
   * @return errorcode.
   */
  protected abstract int closeFile(int fi, int size_hi, int size_lo);

  /**
   * Format filesystem.
   * This will delete all files. If the filesystem integrity is lost,
   * it will try to create a new filesystem. If that fails the error
   * INTEGRITY_LOST is returned, which is a fatal error.
   *
   * @return errorcode.
   *         Returns OK if filesytem formatted ok.
   */
  protected abstract int formatFileSystem();

  /**
   * Get file size (high word).
   * This method internally stores the low word.
   * The file index is valid.
   *
   * @param fi File index.
   * @return High word of filesize or errorcode.
   */
  protected abstract int filesize(int fi);

  /**
   * Get file size (low word).
   * This returns the internally stored low word.
   *
   * @return Low word of filesize.
   */
  protected abstract int filesize();

  /**
   * Write file sector.
   * This method writes the transfer buffer to a physical sector.
   *
   * @param buf Sector buffer.
   * @param fi File index.
   * @param sector Logical file sector.
   * @return errorcode.
   */
  protected abstract int writeSector(char[] buf, int fi, int sector);

  /**
   * Read file sector.
   * This method reads a physical sector into the transfer buffer.
   *
   * @param buf Sector buffer.
   * @param fi File index.
   * @param sector Logical file sector.
   * @return errorcode.
   */
  protected abstract int readSector(char[] buf, int fi, int sector);

  /**
   * Add sector to file.
   *
   * @param fi File index.
   * @return errorcode.
   */
  protected abstract int addSector(int fi);

  /**
   * Remove sector from file.
   *
   * @param fi File index.
   * @return errorcode.
   */
  protected abstract int removeSector(int fi);

  /**
   * Convert hi:lo to sector
   *
   * @param hi High word
   * @param lo Low word
   * @return sector
   */
  protected int toSector(int hi, int lo) {
    return (hi<<(16-SECTORBITS))|(lo>>>SECTORBITS);
  }

  /**
   * Convert hi:lo to offset
   *
   * @param hi High word
   * @param lo Low word
   * @return offset
   */
  protected int toOffset(int hi, int lo) {
    return lo&((1<<SECTORBITS)-1);
  }

  /**
   * Convert sector:offset to hi
   *
   * @param sector Sector
   * @param offset Offset
   * @return hi
   */
  protected int toHi(int sector, int offset) {
    return sector>>>(16-SECTORBITS);
  }

  /**
   * Convert sector:offset to lo
   *
   * @param sector Sector
   * @param offset Offset
   * @return lo
   */
  protected int toLo(int sector, int offset) {
    return (sector<<SECTORBITS)|offset;
  }

}