package stamp.filesystem;
import stamp.peripheral.disk.*;
import stamp.util.text.*;
import stamp.core.*;
import stamp.errors.*;

/**
 * This class extends the AbstractFilesystem16 class and implements a
 * simple and small FAT16 formatted filesystem. This class only accesses
 * files that are listed in the first rootdirectory sector. These files
 * all occupy 32KB of diskspace, starting at the very first cluster.
 * All clusters in a file are consecutive. This means, the entry in the
 * first directory sector corresponds directly to the first cluster of a
 * file.
 *
 * The great advantage of this approach is speed. There is no need to
 * travers the FAT, which could result in loading several sectors just
 * to get to the requested sector. Now the correct sector can be calculated
 * just from the entry number in the rootdirectory sector.
 *
 * To keep the FAT and file entries intact while deleting and creating files,
 * file entries are set to _____NNN.000 for a deleted file, where NNN is the
 * entry number. To create a file, the first entry that matches _____???.???
 * is used and renamed to the new name.
 * In all respects the files with a name with 5 starting underscores are
 * normal files, so they could be opened and written to. Any create() call
 * will however simply rename it and setting its new size to 0, so it is
 * best not to open such files. This can be easily avoided by not using
 * filenames that start with 5 underscores.
 *
 * To have variable filesize, the filename extension field is used
 * to store the actual filesize as used by this class. DOS and Windows
 * will still see a 32KB size. For this reason the filename extension
 * is ignored in the methods that take a filename argument. The created
 * extension is perfectly valid for DOS and Windows and consists only of
 * the digits '0'-'9' and the letters 'A'-'V'. Setting the extension
 * to '000' under DOS or Windows will set the filesize to 0.
 * You can use the CTRL-Z character (0x1A) to define an end of file
 * position inside the file. This is old CPM-style but is still recognized
 * by many DOS/Windows editors. This is however only useful in text files.
 *
 * The normal Windows editors Notepad and Wordpad CANNOT be used to edit
 * the files. These programs make a copy of edited files. This corrupts
 * the FAT for this class.
 * A perfect free editor, called PDT-DOS, is available at
 * http://www.pro-central.com/pdt_dos.htm
 * This editor edits a file ON DISK. The filesectors are simply overwritten
 * and the FAT or filesize is not changed. The program runs under DOS and
 * any Windows version. The free evaluation copy allows to edit the first
 * 25,000 characters in a file. This is sufficient as this class only
 * supports filesizes to 0x7FFE.
 *
 * To create a fresh filesystem, format the disk under DOS or Windows
 * using the FAT16 option. DO NOT USE A LABEL. Then connect the disk (for
 * example CF card) to the Javelin circuit and run SmallFat16SetupCF.java
 * that creates a new filesystem.
 * Another option is, while in DOS or Windows, to copy 32KB files to the
 * disk. Use only 8.3 format filenames. Make sure all extensions are 000.
 * A normal disk has 16 entries in each rootdirectory sector so you
 * should copy 16 files.
 *
 * <p><pre>
 * Revision History:
 * September 20, 2003 v1.0: Initial release.
 * September 26, 2003 v1.1: Added support for multiple primary partitions.
 *                          Updated documentation.
 * September 28, 2003 v1.2: Implemented create() and deleteFile().
 * </pre><p>
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.2 September 28, 2003
 */

public class Fat16FSinterface extends FSinterface {

  public boolean isFat16; //set to true if first partition is valid FAT16
    //the following values are set from partition table
  public int SectorBootRecord_lo = 0;   //start sector of bootrecord
  public int SectorBootRecord_hi = 0;
  public int PartitionSectors_lo = 0;      //number of sectors in partition
  public int PartitionSectors_hi = 0;
  //The follwing values are set from bootrecord
  public int BytesPerSector = 0;
  public int ReservedSectors = 0;
  public int CopiesOfFat = 0;
  public int SectorsPerFat = 0;
  public int SectorFirstFat_lo = 0;
  public int SectorFirstFat_hi = 0;
  public int SectorRootDir_lo = 0;
  public int SectorRootDir_hi = 0;
  public int SectorDataArea_lo = 0;
  public int SectorDataArea_hi = 0;
  public int RootDirEntries = 0;
  public int SectorsPerCluster = 0;
  public int SectorsInRootDir = 0;

  public char[] diskbuf;

  private static final int MAXFILESIZE = 0x7FFE;
  private static final int DIRENTRYSIZE = 32;

  //parameters for 32K filesize
  private int clusters32K; //number of clusters for 32K filesize
  private int fileClusters;
  private int lastsearch = 0;

  //shift bits value for cluster/sector conversion
  private int SectorClusterConversionShift;

  //dir sector LBA
  private int dirsector_lo; //result of dirSector(fi)
  private int dirsector_hi;
  private int diroffset;

  //fat sector LBA
  private int fatsector_lo; //result of fatSector(cluster)
  private int fatsector_hi;
  private int fatoffset;

  //file sector LBA
  private int filesector_lo; //result of fileSector(fi,sector)
  private int filesector_hi;

  //current read/write sector LBA
  private int cursector_lo; //current sector in diskbuf
  private int cursector_hi;

  private Disk disk; //provides access to storage media
  private int filesize_lo;
  private char[] filename = new char[13];

  private int firstCluster;
  private int forelastCluster;
  private int lastCluster;
  private int fatSector_lo;
  private int fatSector_hi;

  /**
   * Constructor that defines a small FAT16 filesystem.
   * It imports the sectorsize and its power of 2 value from the
   * disk device.
   *
   * @param disk The disk device of the filesystem.
   * @param partition The partition (0-3) on the disk where the filesystem resides.
   * @param openFiles The maximum number of simultaneously open files.
   * @param buf The sector transfer buffer.
   */
  public Fat16FSinterface(Disk disk) {
    super(disk.SECTORSIZE);
    this.disk = disk;
    this.diskbuf = new char[disk.SECTORSIZE];
    filename[12] = 0; //asciiz
  }

  /**
   * Read parameters from disk.
   * The first partition must be a valid FAT16 partition with
   * 512 bytes per sector. Only then true will be returned.
   *
   * @return True if first partition is valid FAT16 partition
   *         with 512 bytes per sector.
   */
  public boolean readDiskParameters() {
    return isFat16 = init();
  }

  /**
   * Read parameters from disk.
   * The first partition must be a valid FAT16 partition with
   * 512 bytes per sector. Only then true will be returned.
   *
   * @return True if first partition is valid FAT16 partition
   *         with 512 bytes per sector.
   */
  private boolean init() {
    if (!disk.present()) return false;
    while (!disk.ready()) ;
    //read masterbootrecord
    disk.readSector(diskbuf,0,0);
    //check disk validity
    //if word at offset 0x1FE != 0xAA55 the disk is invalid
    if (readWord(0x1FE) != (short)0xAA55) return false;
    //check partition table entry 0
    int offset = 0x1BE;
    //test if boot indicator is either 0x80 (active) or 0x00 (inactive)
    if ((diskbuf[offset] & 0x7F) != 0x00) return false;
    //get partition type
    int type = diskbuf[offset+4]&0xFF;
    if ((type != 0x04) && (type != 0x06)) return false;
    //get the start sector for the partition
    SectorBootRecord_lo = readWord(offset+8);
    SectorBootRecord_hi = readWord(offset+10);
    //get the number of sectors
    PartitionSectors_lo = readWord(offset+12);
    PartitionSectors_hi = readWord(offset+14);
    //read bootrecord
    disk.readSector(diskbuf,SectorBootRecord_hi,SectorBootRecord_lo);
    //check bootrecord for validity
    if (readWord(0x1FE) != (short)0xAA55) return false; //invalid marker
    //get the bytes per sector field
    BytesPerSector = readWord(11);
    //get the sectors per cluster field
    SectorsPerCluster = diskbuf[13]&0xFF;
    //get the reserved sectors
    ReservedSectors = readWord(14);
    //get the copies of fat
    CopiesOfFat = diskbuf[16]&0xFF;
    //get the max number of root entries
    RootDirEntries = readWord(17);
    //get the sectors per fat
    SectorsPerFat = readWord(22);
    //start by adding the starting sector with the reserved sectors
    SectorFirstFat_hi = SectorBootRecord_hi;
    SectorFirstFat_lo = SectorBootRecord_lo + ReservedSectors;
    if (CPU.carry()) SectorFirstFat_hi++;
    //add the total sectors in all copies of the FAT table
    SectorRootDir_lo = SectorsPerFat * CopiesOfFat;
    SectorRootDir_lo += SectorFirstFat_lo;
    if (CPU.carry()) SectorRootDir_hi++;
    //calculate the data sector offset based on the max entries * 32 (32 bytes each)
    //and divide the offset by bytes per sector to determine the sector number
    SectorsInRootDir = (RootDirEntries * DIRENTRYSIZE) / BytesPerSector;
    SectorDataArea_hi = SectorRootDir_hi;
    SectorDataArea_lo = SectorRootDir_lo + SectorsInRootDir;
    if (CPU.carry()) SectorDataArea_hi++;
    return true;
  }

  /*
  Master Boot Record
  Offset  Description                               Size
  000h    Executable Code (Boots Computer)        446 Bytes
  1BEh    1st Partition Entry                      16 Bytes
  1CEh    2nd Partition Entry                      16 Bytes
  1DEh    3rd Partition Entry                      16 Bytes
  1EEh    4th Partition Entry                      16 Bytes
  1FEh    Executable Marker (55h AAh)               2 Bytes

  Partition entry
  Offset  Description                               Size
  00h     Boot indicator (80h/00h)                  1 Byte
  01h     Beginning of Partition - Head             1 Byte
  02h     Beginning of Partition - Cylinder/Sector  1 Word
  04h     Type of Partition                         1 Byte
  05h     End of Partition - Head                   1 Byte
  06h     End of Partition - Cylinder/Sector        1 Word
  08h     Number of Sectors Between the MBR and
          the First Sector in the Partition         1 Double Word
  0Ch     Number of Sectors in the Partition        1 Double Word

  Partition type
  00h     Unknown or Nothing
  01h     12-bit FAT
  04h     16-bit FAT (Partition Smaller than 32MB)
  05h     Extended (MS-DOS) Partition
  06h     16-bit FAT (Partition Larger than 32MB)
  0Bh     32-bit FAT (Partition Up to 2048GB)
  0Ch     Same as 0BH, but uses LBA 13h Extensions
  0Eh     Same as 06H, but uses LBA 13h Extensions
  0Fh     Same as 05H, but uses LBA 13h Extensions

  Cylinder/sector
  15 14 13 12 11 10 9 8   7 6                 5 4 3 2 1 0
  Cylinder Bits 7 to 0    Cylinder Bits 9+8   Sector Bits 5 to 0
  */

  /*
  FAT16 Boot Record
  This information is located in the first sector of every partition.
  Offset  Description                                       Size
  00h     Jump Code + NOP                                   3 Bytes
  03h     OEM Name                                          8 Bytes
  0Bh     Bytes Per Sector                                  1 Word
  0Dh     Sectors Per Cluster                               1 Byte
  0Eh     Reserved Sectors                                  1 Word
  10h     Number of Copies of FAT                           1 Byte
  11h     Maximum Root Directory Entries                    1 Word
  13h     Number of Sectors in Partition Smaller than 32MB  1 Word
  15h     Media Descriptor (F8h for Hard Disks)             1 Byte
  16h     Sectors Per FAT                                   1 Word
  18h     Sectors Per Track                                 1 Word
  1Ah     Number of Heads                                   1 Word
  1Ch     Number of Hidden Sectors in Partition             1 Double Word
  20h     Number of Sectors in Partition                    1 Double Word
  24h     Logical Drive Number of Partition                 1 Word
  26h     Extended Signature (29h)                          1 Byte
  27h     Serial Number of Partition                        1 Double Word
  2Bh     Volume Name of Partition                          11 Bytes
  36h     FAT Name (FAT16)                                  8 Bytes
  3Eh     Executable Code                                   448 Bytes
  1FEh    Executable Marker (55h AAh)                       2 Bytes

  FAT16 Drive Layout
  Offset                                                          Description
  Start of Partition                                              Boot Sector
  Start + # of Reserved Sectors                                   Fat Tables
  Start + # of Reserved + (# of Sectors Per FAT * 2)              Root Directory Entry
  Start + # of Reserved +
    (# of Sectors Per FAT * 2) +
    ((Maximum Root Directory Entries * 32) / Bytes per Sector)    Data Area (Starts with Cluster #2)
  */

  /**
   * Test if device is present.
   * If the device does not support this it must return true.
   *
   * @return True if device is present.
   */
  protected boolean present() {
    return disk.present();
  }

  /**
   * Test if device is ready.
   * If the device does not support this it must return true.
   *
   * @return True if device is ready.
   */
  protected boolean ready() {
    return disk.ready();
  }

  /**
   * Reset device.
   * If the device does not support this it must return true.
   *
   * @return True if reset succesful.
   */
  protected boolean reset() {
    return disk.reset();
  }

  /**
   * Get device status.
   * If the device does not support this it must return 0.
   *
   * @return Errorcode.
   */
  protected int status() {
    return disk.status();
  }

  /**
   * Read file sector.
   * This method reads a physical sector into the transfer buffer.
   *
   * @param buf Sector buffer.
   * @param fi File index.
   * @param sector Logical file sector.
   * @return errorcode.
   */
  protected int readSector(char[] buf, int fi, int sector) {
    setFileSectorLBA(fi,sector);
    return readAbsSector(buf,filesector_hi,filesector_lo);
  }

  /**
   * Write file sector.
   * This method writes the transfer buffer to a physical sector.
   *
   * @param buf Sector buffer.
   * @param fi File index.
   * @param sector Logical file sector.
   * @return errorcode.
   */
  protected int writeSector(char[] buf, int fi, int sector) {
    setFileSectorLBA(fi,sector);
    return writeAbsSector(buf,filesector_hi,filesector_lo);
  }

  /**
   * Get number of files.
   *
   * @return Number of files.
   */
  protected int files() {
    int files = 0;
    int sector_lo = SectorRootDir_lo;
    int sector_hi = SectorRootDir_hi;
    int entries = SECTORSIZE / DIRENTRYSIZE;
    int offset,entry,sectors;
    for (sectors=0; sectors<SectorsInRootDir; sectors++) {
      readAbsSector(diskbuf,sector_hi,sector_lo);
      for (entry=0; entry<entries; entry++) {
        offset = entry*DIRENTRYSIZE;
        if ((diskbuf[offset] != 0) && (diskbuf[offset] != 0xE5)) { //used entry
          if ((diskbuf[offset+11]&0x18) == 0) files++; //not volume or directory
        }
      }
      sector_lo += 1;
      if (CPU.carry()) sector_hi++;
    }
    return files;
  }

  /**
   * Search a file.
   *
   * @param name Path/name of the file.
   *             See extended class for filename specifications.
   * @return File index or errorcode.
   *         This file index identifies a file on a physical media.
   *         The transferring to/from media is handled by the extended class.
   */
  protected int searchFile(String name) {
  /*
    //directory entry layout (32 bytes)
    char name[8];            0     // ALL-CAPS, pad right with spaces
    char ext[3];            +8     // ALL-CAPS, pad right with spaces, we use this to store actual size
    char attrib;           +11     // attribute byte
    char reserved;         +12     // =0
    char ctime_ms;         +13     // file creation time, 10ms units
    int  ctime;            +14     // file creation time, in DOS format
    int  cdate;            +16     // file creation date, in DOS format
    int  adate;            +18     // DOS date of last file access
    int  cluster_fat32;    +20     // high 16 bits of starting cluster (FAT32, 0 for FAT16)
    int  mtime;            +22     // DOS time of last file modification
    int  mdate;            +24     // DOS date of last file modification
    int  cluster;          +26     // starting cluster
    int  file_size_lo;     +28     // lowest 16 bits of filesize in bytes
    int  file_size_hi;     +30     // highest 16 bits of filesize in bytes

    //DOS time format (16 bits)
    bits   meaning   values
    4-0    2-second  0-29
    10-5   minutes   0-59
    15-11  hours     0-23

    //DOS date format (16 bits)
    bits   meaning           values
    4-0    day               1-31
    8-5    month             1-12
    15-9   year since 1980   0-127

    //attribute = --ADVSHR
    A = archive
    D = directory  if set defines directory, V=0 and filesize=0
    V = volume     if set defines volume, other bits 0, filesize 0
    S = system     if set defines system
    H = hidden     if set defines hidden
    R = read only  if set defines read only
    for a file: dont cares are -,A,S,H,R
                must be clear D,V
    */
    int cntr,entry,offset,sectors;
    int sector_lo = SectorRootDir_lo;
    int sector_hi = SectorRootDir_hi;
    int entries = SECTORSIZE/DIRENTRYSIZE;
    for (sectors=0; sectors<SectorsInRootDir; sectors++) {
      readAbsSector(diskbuf,sector_hi,sector_lo);
      for (entry=0; entry<entries; entry++) {
        offset = entry*DIRENTRYSIZE;
        for (cntr=0; cntr<8; cntr++) {
          if (name.charAt(cntr) != diskbuf[offset+cntr]) break;
        }
        if (cntr == 8) { //check extension
          if (name.charAt(8) != '.') return err.NOT_FOUND;
          for (cntr=8; cntr<11; cntr++) {
            if (name.charAt(cntr+1) != diskbuf[offset+cntr]) break;
          }
          if (cntr == 11) {
            if ((diskbuf[offset+11]&0x18) != 0) return err.ILLEGAL; //volume or directory
            return (sectors*(SECTORSIZE/DIRENTRYSIZE))+entry+1; //file index >= 1
          }
        }
      }
      sector_lo += 1;
      if (CPU.carry()) sector_hi++;
    }
    return err.NOT_FOUND;
  }

  /**
   * Format filesystem.
   * This will delete all files. If the filesystem integrity is lost,
   * it will try to create a new filesystem. If that fails the error
   * INTEGRITY_LOST is returned, which is a fatal error.
   *
   * @return errorcode.
   *         Returns OK if filesytem formatted ok.
   */
  protected int formatFileSystem() {
    return err.NOT_IMPLEMENTED;
  }

  /**
   * Add sector to file.
   *
   * @param fi File index.
   * @return errorcode.
   */
  protected int addSector(int fi) {
    findFileClusters(fi);
    int newCluster = findFreeCluster();
    if (!dirSector(fi-1)) return err.INVALID;
    readAbsSector(diskbuf,dirsector_hi,dirsector_lo); //read rootdir sector holding entry
    int size_lo = readWord(diroffset+28);
    int size_hi = readWord(diroffset+30);
    int firstcluster = readWord(diroffset+26);
    int sectors = (size_hi<<(16-SECTORBITS))|(size_lo>>>SECTORBITS);
    if ((size_lo & (SECTORSIZE-1)) != 0) sectors++; //current sectors
    sectors++;
    if ((sectors % SectorsPerCluster) == 1) { //must add cluster
      if (firstcluster == 0) {
        writeWord(diroffset+26,newCluster);
        writeAbsSector(diskbuf,dirsector_hi,dirsector_lo);
      }
    }
    else {
      int next = 0;
      int cluster = firstcluster;
      while ((next>>>3) != 0x1FFF) {
        if (!fatSector(cluster)) return err.NOT_FOUND;
        readAbsSector(diskbuf,fatsector_hi,fatsector_lo);
        next = readWord(fatoffset);
      }
      //cluster is now last cluster, fatoffset is entry of last cluster
      writeWord(fatoffset,newCluster);
      writeAbsSector(diskbuf,fatsector_hi,fatsector_lo);
    }
    fatSector(newCluster);
    readAbsSector(diskbuf,fatsector_hi,fatsector_lo);
    writeWord(fatoffset,(short)0xFFF8);
    writeAbsSector(diskbuf,fatsector_hi,fatsector_lo);
    return err.OK;
  }

  /**
   * Calculate DIR sector for entry.
   * dirsector_hi and dirsector_lo set to sector holding directory entry in rootdir
   * diroffset set to offset in that sector
   *
   * @param entry Entry number.
   * @return true if entry value valid
   */
  private boolean dirSector(int entry) {
    int maxrootentry = SectorsInRootDir * (SECTORSIZE/DIRENTRYSIZE);
    if (entry - maxrootentry >= 0) return false; //entry too high
    int dirsector = entry/DIRENTRYSIZE;
    dirsector_hi = SectorRootDir_hi;
    dirsector_lo = SectorRootDir_lo + dirsector;
    if (CPU.carry()) dirsector_hi++;
    diroffset = (entry % (SECTORSIZE/DIRENTRYSIZE)) * DIRENTRYSIZE;
    return true;
  }

  /**
   * Calculate FAT sector for cluster.
   * fatsector_hi and fatsector_lo set to sector holding cluster entry in FAT
   * fatoffset set to offset in that sector
   *
   * @param cluster Cluster number.
   * @return true if cluster value valid
   */
  private boolean fatSector(int cluster) {
    int maxcluster = SectorsPerFat << (SECTORBITS-1); //equals SectorsPerFat*(SECTORSIZE/2);
    if (cluster - maxcluster >= 0) return false; //cluster too high
    int fatsector = cluster >>> (SECTORBITS-1); //equals cluster/(SECTORSIZE/2);
    fatsector_hi = SectorFirstFat_hi;
    fatsector_lo = SectorFirstFat_lo + fatsector;
    if (CPU.carry()) fatsector_hi++;
    fatoffset = (cluster % (SECTORSIZE/2)) * 2;
    return true;
  }

  /**
   * Remove sector from file.
   *
   * @param fi File index.
   * @return errorcode.
   */
  protected int removeSector(int fi) {
    readAbsSector(diskbuf,SectorRootDir_hi,SectorRootDir_lo);
    int offset = (fi-1)*DIRENTRYSIZE;
    int size_lo = readWord(offset+28);
    int size_hi = readWord(offset+30);
    int sectors = (size_hi<<(16-SECTORBITS))|(size_lo>>>SECTORBITS);
    if ((size_lo & (SECTORSIZE-1)) != 0) sectors++; //current sectors
    sectors--;
    if ((sectors % SectorsPerCluster) == 0) { //must remove cluster
    }
    return err.OK;
  }

  /**
   * Delete a file.
   * The file index is valid.
   *
   * @param fi File index of file to delete.
   * @return errorcode.
   */
  protected int deleteFile(int fi) {
    int offset,cluster,count;
    //To keep the FAT and file entries intact we simply rename the file
    //to _____NNN.000 where NNN is the entry number. The extension 000
    //defines a size of 0. The create() method searches for any file entry
    //that matches _____??? and uses its entry to create a new file.
    cluster = ((fi-1)*clusters32K) + 2; //start cluster
    offset = (fi-1)*DIRENTRYSIZE; //32 bytes per entry
    //set filename field (_____XXX000)
    //note that extension 000 sets the actual filesize to 0
    Format.bprintf(diskbuf,offset,"_____%03d000",fi-1);
    //set attribute field to archived
    diskbuf[offset+11] = (char)0x20;
    //clear date and time fields
    for (count=12; count<26; count++) diskbuf[offset+count] = 0;
    //set first cluster field
    diskbuf[offset+26] = (char)(cluster&0xFF);
    diskbuf[offset+27] = (char)(cluster>>>8);
    //set filesize field
    diskbuf[offset+28] = (char)(MAXFILESIZE&0xFF);
    diskbuf[offset+29] = (char)(MAXFILESIZE>>>8);
    diskbuf[offset+30] = (char)0x00;
    diskbuf[offset+31] = (char)0x00;
    return err.OK;
  }

  /**
   * Create a file.
   * File size is set to 0.
   *
   * @param name Name of the file in 8.3 format.
   *        The filename extension is ignored.
   *        Do not use invalid DOS filename characters.
   *        Use only capitals and digits to be safe.
   * @return errorcode.
   */
  protected int createFile(String name) {
    int fi,offset,cluster,count;
    fi = searchFile(name);
    if (fi > 0) return err.FOUND; //file exists
    fi = findFreeRootEntry();
    if (fi == 0) return err.OUT_OF_MEMORY; //disk full
    cluster = ((fi-1)*fileClusters) + 2; //start cluster
    offset = (fi-1)*DIRENTRYSIZE; //32 bytes per entry
    //set filename field to new name with extension 000
    //note that extension 000 sets the actual filesize to 0
    for (count=0; count<8; count++) diskbuf[offset+count] = name.charAt(count);
    for (count=8; count<11; count++) diskbuf[offset+count] = '0';
    //set attribute field to archived
    diskbuf[offset+11] = (char)0x20;
    //clear date and time fields
    for (count=12; count<26; count++) diskbuf[offset+count] = 0;
    //set first cluster field
    diskbuf[offset+26] = (char)(cluster&0xFF);
    diskbuf[offset+27] = (char)(cluster>>>8);
    //set filesize field
    diskbuf[offset+28] = (char)(MAXFILESIZE&0xFF);
    diskbuf[offset+29] = (char)(MAXFILESIZE>>>8);
    diskbuf[offset+30] = (char)0x00;
    diskbuf[offset+31] = (char)0x00;
    return err.OK;
  }

  /**
   * Get file size (high word).
   * This method internally stores the low word.
   * The file index is valid.
   *
   * @param fi File index.
   * @return High word of filesize or errorcode.
   */
  protected int filesize(int fi) {
    int offset;
    readAbsSector(diskbuf,SectorRootDir_hi,SectorRootDir_lo);
    offset = (fi-1)*DIRENTRYSIZE;
    filesize_lo = readWord(offset+28);
    return readWord(offset+30);
  }

  /**
   * Get file size (low word).
   * This returns the internally stored low word.
   *
   * @return Low word of filesize.
   */
  protected int filesize() {
    return filesize_lo;
  }

  /**
   * Close a file.
   * The new size will be written to disk
   * The file index is valid.
   *
   * @param fi File index of file to close.
   * @param size_hi High word of new filesize
   * @param size_lo Low word of new filesize
   * @return errorcode.
   */
  protected int closeFile(int fi, int size_hi, int size_lo) {
    int error,offset;
    readAbsSector(diskbuf,SectorRootDir_hi,SectorRootDir_lo);
    offset = (fi-1)*DIRENTRYSIZE;
    writeWord(offset+28,size_lo);
    writeWord(offset+30,size_hi);
    //save sector
    error = writeAbsSector(diskbuf,SectorRootDir_hi,SectorRootDir_lo);
    return (error < 0) ? error : size_lo;
  }

  /**
   * Get filename.
   *
   * @param index Logical entry for file.
   *              Range is 0 to files()-1.
   * @return Name of file as ASCIIZ or null.
   */
  protected char[] fileName(int index) {
    int i,offset,entry;
    readAbsSector(diskbuf,SectorRootDir_hi,SectorRootDir_lo);
    int entries = SECTORSIZE/DIRENTRYSIZE;
    for (entry=0; entry<entries; entry++) {
      offset = entry*DIRENTRYSIZE;
      if ((diskbuf[offset] != 0) && (diskbuf[offset] != 0xE5)) { //used entry
        if ((diskbuf[offset+11]&0x18) == 0) { //not volume or directory
          if (index == 0) { //this is the entry
            for (i=0; i<8; i++) filename[i] = diskbuf[offset+i];
            filename[8] = '.';
            for (i=8; i<11; i++) filename[i+1] = diskbuf[offset+i];
            return filename;
          }
          else index--; //check next file
        }
      }
    }
    return null;
  }

  /**
   * Find free cluster.
   *
   * @param fi File index.
   * @return errorcode.
   */
  private int findFreeCluster() {
    return 0;
  }

  /**
   * Find important clusters for file.
   * (first, forelast and last)
   *
   * @param fi File index.
   */
  private void findFileClusters(int fi) {
    int cluster;
    //get first cluster for file
    readAbsSector(diskbuf,SectorRootDir_hi,SectorRootDir_lo);
    int offset = (fi-1)*DIRENTRYSIZE;
    firstCluster = readWord(offset+26);
    forelastCluster = 0;
    lastCluster = 0;
    if (firstCluster > 0) {
      cluster = firstCluster;
      int ClusterEntriesPerSector = SECTORSIZE/2;
      while ((cluster&(short)0xFFF8)!=(short)0xFFF8) { //not last cluster
        //convert cluster to FAT sector that holds cluster entry
        int lfatsector = cluster / ClusterEntriesPerSector;
        //calculate fat sector
        int sector_lo = SectorFirstFat_lo;
        int sector_hi = SectorFirstFat_hi;
        sector_lo += lfatsector;
        if (CPU.carry()) sector_hi++;
        //get next cluster from chain
        readAbsSector(diskbuf,sector_hi,sector_lo);
        int lfatentry = (cluster % ClusterEntriesPerSector) << 1; //wordentry
        forelastCluster = lastCluster;
        lastCluster = cluster;
        cluster = readWord(lfatentry);
      }
    }
  }

  /**
   * Find free rootdirectory entry.
   *
   * @return File index or 0.
   */
  private int findFreeRootEntry() {
    int i,cntr,entry,offset,fi;
    int sector_lo = SectorRootDir_lo;
    int sector_hi = SectorRootDir_hi;
    int entries = SECTORSIZE/DIRENTRYSIZE;
    for (i=0; i<SectorsInRootDir; i++) {
      readAbsSector(diskbuf,sector_hi,sector_lo);
      for (entry=0; entry<entries; entry++) {
        offset = entry*DIRENTRYSIZE;
        fi = entry+(entries*i)+1;
        if ((diskbuf[offset] == 0) || ((diskbuf[offset]&0xFF) == 0xE5)) return fi;
      }
      sector_lo += 1;
      if (CPU.carry()) sector_hi++;
    }
    return 0;
  }

  /**
   * Calculate LBA for logical file sector.
   * This method sets the globals filesector_lo and filesector_hi.
   *
   * @param fi File index.
   * @param sector Logical file sector.
   */
  private void setFileSectorLBA(int fi, int sector) {
    int sector_lo, sector_hi;
    //convert logical sector to logical cluster
    int lcluster = sector / SectorsPerCluster;
    //get first cluster for file
    readAbsSector(diskbuf,dirsector_hi,dirsector_lo);
    int offset = ((fi-1)%(SECTORSIZE/DIRENTRYSIZE))*DIRENTRYSIZE;
    int cluster = readWord(offset+26);
    int ClusterEntriesPerSector = SECTORSIZE/2;
    while (lcluster > 0) {
      readAbsSector(diskbuf,fatsector_hi,fatsector_lo);
      //get next cluster from chain
      int offset2 = (cluster % ClusterEntriesPerSector) << 1; //wordentry
      cluster = readWord(offset2);
      lcluster--;
    }
    filesector_hi = (cluster-2) >>> (16-SectorClusterConversionShift);
    filesector_lo = (cluster-2) << SectorClusterConversionShift;
    filesector_lo += SectorDataArea_lo;
    if (CPU.carry()) filesector_hi++;
    filesector_hi += SectorDataArea_hi;
    if (CPU.carry()) filesector_hi++;
  }

  /**
   * Read absolute sector.
   * This method reads a physical sector into the transfer buffer.
   *
   * @param buf Sector buffer.
   * @param sector_hi High word of sector.
   * @param sector_lo Low word of sector.
   * @return errorcode.
   */
  private int readAbsSector(char[] buf, int sector_hi, int sector_lo) {
    if ((cursector_lo != sector_lo) || (cursector_hi != sector_hi)) {
      cursector_lo = sector_lo;
      cursector_hi = sector_hi;
      if (disk.readSector(buf,sector_hi,sector_lo)) return err.OK;
      return err.OUT_OF_RANGE;
    }
    return err.OK;
  }

  /**
   * Write absolute sector.
   * This method writes the transfer buffer to a physical sector.
   *
   * @param buf Sector buffer.
   * @param sector_hi High word of sector.
   * @param sector_lo Low word of sector.
   * @return errorcode.
   */
  private int writeAbsSector(char[] buf, int sector_hi, int sector_lo) {
    cursector_lo = sector_lo;
    cursector_hi = sector_hi;
    if (disk.writeSector(buf,sector_hi,sector_lo)) return err.OK;
    return err.OUT_OF_RANGE;
  }

  /**
   * Get word from diskbuf.
   *
   * @param addr Position in diskbuf.
   * @return word at position.
   */
  private int readWord(int addr) {
    return (diskbuf[addr]&0xFF)|(diskbuf[addr+1]<<8);
  }

  /**
   * Write word to diskbuf.
   *
   * @param addr Position in diskbuf.
   * @param value Word to write.
   */
  private void writeWord(int addr, int value) {
    diskbuf[addr++] = (char)(value&0xFF);
    diskbuf[addr] = (char)(value>>>8);
  }

  /**
   * Print all filenames + filesizes to Javelin message window.
   */
  public void listFiles() {
    int i,fi,size,dig,offset;
    fi = -1;
    do {
      fi = search("????????.???",fi+1);
      if (fi != 0) {
        offset = (fi-1) * DIRENTRYSIZE;
        for (i=0; i<8; i++) Format.printf("%c",diskbuf[offset+i]);
        Format.printf(".");
        for (i=8; i<11; i++) Format.printf("%c",diskbuf[offset+i]);
        int size_lo = readWord(offset+28);
        int size_hi = readWord(offset+30);
        Format.printf("  %04x",size_hi);
        Format.printf("%04x bytes\n",size_lo);
      }
    } while (fi != 0);
  }

  public int search(String name, int start) {
  /*
    //directory entry layout (32 bytes)
    char name[8];            0     // ALL-CAPS, pad right with spaces
    char ext[3];            +8     // ALL-CAPS, pad right with spaces, we use this to store actual size
    char attrib;           +11     // attribute byte
    char reserved;         +12     // =0
    char ctime_ms;         +13     // file creation time, 10ms units
    int  ctime;            +14     // file creation time, in DOS format
    int  cdate;            +16     // file creation date, in DOS format
    int  adate;            +18     // DOS date of last file access
    int  cluster_fat32;    +20     // high 16 bits of starting cluster (FAT32, 0 for FAT16)
    int  mtime;            +22     // DOS time of last file modification
    int  mdate;            +24     // DOS date of last file modification
    int  cluster;          +26     // starting cluster
    int  file_size_lo;     +28     // lowest 16 bits of filesize in bytes
    int  file_size_hi;     +30     // highest 16 bits of filesize in bytes

    //DOS time format (16 bits)
    bits   meaning   values
    4-0    2-second  0-29
    10-5   minutes   0-59
    15-11  hours     0-23

    //DOS date format (16 bits)
    bits   meaning           values
    4-0    day               1-31
    8-5    month             1-12
    15-9   year since 1980   0-127

    //attribute = --ADVSHR
    A = archive
    D = directory  if set defines directory, V=0 and filesize=0
    V = volume     if set defines volume, other bits 0, filesize 0
    S = system     if set defines system
    H = hidden     if set defines hidden
    R = read only  if set defines read only
    for a file: dont cares are -,A,S,H,R
                must be clear D,V
    */
    int cntr,entry,offset;
    //check for previous search
    if ((lastsearch+1) == start) { //get index of previous search
      start--;
    }
    else { //start search from first file entry
      start = 0;
    }
    readAbsSector(diskbuf,SectorRootDir_hi,SectorRootDir_lo);
    for (entry=start; entry<(SECTORSIZE/DIRENTRYSIZE); entry++) {
      offset = entry*DIRENTRYSIZE;
      int c = diskbuf[offset]&0xFF;
      if ((c!=0) && (c!=0xE5)) { //used entry
        //LFN entries have Read-only, System, Hidden, Volume set
        int attr = diskbuf[offset+11]&0xFF;
        if ((attr & 0x0F) != 0x0F) { //not LFN entry
          if ((attr&0x18) == 0) { //not volume or directory
            lastsearch = entry+1; //file index >= 1
            return lastsearch;
          }
        }
      }
    }
    lastsearch=0;
    return 0;
  }

}