import stamp.peripheral.memory.compactflash.*;
import stamp.errors.*;
import stamp.filesystem.*;
import stamp.core.*;
import stamp.peripheral.disk.*;
import stamp.protocol.*;
import stamp.util.text.*;

/**
 * This class tests the FileSystem class with a Fat16FSinterface object.
 *
 * <p><pre>
 * Revision History:
 * September 13, 2004 v1.0: Initial release.
 * </pre><p>
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 September 13, 2004
 */

public class Fat16FileSystem_test {

  static char[] sectorbuf = new char[512];

  //i2cbus
  static final int SDA  = CPU.pin8;
  static final int SCL  = CPU.pin9;
  //create objects that control a cf card
  static I2CsmFastDevice CF_DATA = new I2CsmFastDevice(SDA,SCL,0x70,32767,1,1); //PCF8574A
  static I2CsmFastDevice CF_CONTROL = new I2CsmFastDevice(SDA,SCL,0x42,32767,1,1); //PCF8574
  //create I2C interface for cf card
  static I2C_CompactFlash cfi = new I2C_CompactFlash(CF_DATA,CF_CONTROL);
  //create cf object
  static CompactFlash cf = new CompactFlash(cfi);
  //create FAT16 filesystem interface for cf
  static Fat16FSinterface fsi = new Fat16FSinterface(cf);
  //create filesystem
  static FileSystem fs = new FileSystem(fsi,10);

  static final String A = new String("FILE_433.TXT");
  static final String B = new String("FILE_032.TXT");
  static char[] text = new char[512];

  /**
   * Print info about compact flash card.
   */
  static void printDeviceParameters()                        /* Display information of the        */
  {                                        /* CompactFlash Card                */
    Format.printf("Model number                : %s\n",cf.Model);
    Format.printf("Firmware revision           : %s\n",cf.Firmware);
    Format.printf("Serial number               : %s\n",cf.Serial);
    Format.printf("Default cylinders           : %u\n",cf.DefaultCylinders);
    Format.printf("Default heads               : %u\n",cf.DefaultHeads);
    Format.printf("Default sectors per track   : %u\n",cf.DefaultSectors);
    Format.printf("Current cylinders           : %u\n",cf.CurrentCylinders);
    Format.printf("Current heads               : %u\n",cf.CurrentHeads);
    Format.printf("Current sectors per track   : %u\n",cf.CurrentSectors);
    Format.printf("Unformatted bytes per sector: %u\n",cf.UnformattedBytes);
    Format.printf("Current capacity LBA_lo     : %u\n",cf.CurrentCapacity_lo);
    Format.printf("Current capacity LBA_hi     : %u\n",cf.CurrentCapacity_hi);
    Format.printf("Addressable sectors LBA_lo  : %u\n",cf.AddressableSectors_lo);
    Format.printf("Addressable sectors LBA_hi  : %u\n\n",cf.AddressableSectors_hi);
  }

  /**
   * Print the disk related parameters
   */
  static void printDiskParameters() {
    Format.printf("Start of boot record: %04x",fsi.SectorBootRecord_hi);
    Format.printf("%04xh\r\n",fsi.SectorBootRecord_lo);
    Format.printf("Start of first FAT  : %04x",fsi.SectorFirstFat_hi);
    Format.printf("%04xh\r\n",fsi.SectorFirstFat_lo);
    Format.printf("Start of root dir   : %04x",fsi.SectorRootDir_hi);
    Format.printf("%04xh\r\n",fsi.SectorRootDir_lo);
    Format.printf("Start of data area  : %04x",fsi.SectorDataArea_hi);
    Format.printf("%04xh\r\n\r\n",fsi.SectorDataArea_lo);
    Format.printf("Bytes per sector: %d\r\n",fsi.BytesPerSector);
    Format.printf("Sectors per cluster: %d\r\n",fsi.SectorsPerCluster);
    Format.printf("Sectors per fat: %d\r\n",fsi.SectorsPerFat);
    Format.printf("Reserved sectors: %d\r\n",fsi.ReservedSectors);
    Format.printf("Copies of fat: %d\r\n",fsi.CopiesOfFat);
    Format.printf("Max root entries: %d\r\n",fsi.RootDirEntries);
  }

  static void main() {
    int i,j,k,fd=0,error,fe;
    Format.printf("Fat16FileSystem test program\n");
    Format.printf("Memory free: %u\n",stamp.core.Memory.freeMemory());

    if (!cfi.present()) {
      Format.printf("CF card not inserted\n");
      Format.printf("Program halted\n");
      while (true) ;
    }
    else {
      Format.printf("CF card reset ");
      if (cf.reset()) {
        Format.printf("OK\n");
      }
      else {
        Format.printf("failed!\n");
        Format.printf("Program halted\n");
        while (true) ;
      }
    }

    Format.printf("Flashing the CFIO led\n");
    for (i=0; i<6; i++) {
      cfi.enable();
      for (k=0; k<6000; k++) ;
      cfi.disable();
      for (k=0; k<6000; k++) ;
    }

    //the following should be done before reading disk parameters
    Format.printf("Reading device information...\n");
    cf.readDeviceParameters(sectorbuf);
    printDeviceParameters();

    //the following MUST be done before using filesystem
    Format.printf("Reading disk parameters...\n");
    if (!fsi.readDiskParameters()) {
      Format.printf("Disk not FAT16 formatted\n");
      Format.printf("Program halted\n");
      while (true) ;
    }
    printDiskParameters();

/*
    if (fs.status() != err.OK) { //try to create filesystem
      if (fs.format() != err.OK) {
        Format.printf("Unable to create filesystem\n");
        Format.printf("Program halted\n");
        while (true) ;
      }
    }
*/
    //fsi.listFiles();

/*
    //create file.
    Format.printf("Try to create file %s\n",A);
    fd = fs.open(A,fs.APPEND);
    if (fd <= 0) printError(fd);
    else {
      for (i=0; i<26; i++) {
        Format.printf("write line of 80 chars + newline = 81 bytes\n");
        for (k=0; k<8; k++) {
          for (j=0; j<10; j++) {
            fs.writeByte(fd,'0'+j);
          }
        }
        fs.writeByte(fd,10); //newline
        fs.flush(fd);
        //fsi.listFiles();

        //read file
        Format.printf("Read file so far\n");
        fe = fs.open(A,fs.READ);
        if (fe <= 0) printError(fe);
        else {
          while ((k=fs.readByte(fe)) > 0) Format.printf("%c",k);
          printError(k); //should be end of file
          fs.close(fe);
        }
      }
      fs.tell(fd); //get high word of filesize (=0)
      Format.printf("filesize: %d\n",fs.tell()); //should be 100 * 31 = 3100
      fs.close(fd);
    }
*/
/*    //append to file
    Format.printf("Try to append to file %s\n",A);
    fd = fs.open(A,fs.APPEND);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("appending 3 lines ABCDEF + newline = 21 bytes\n");
      for (i=0; i<3; i++) {    // 3 lines of "ABCDEF" + newline
        for (j=0; j<6; j++) {
          fs.writeByte(fd,'A'+j);
        }
        fs.writeByte(fd,10); //newline
      }
      fs.seek(fd,0,0,fs.SEEK_END);
      fs.tell(fd); //get high word of filesize (=0)
      Format.printf("filesize: %d\n",fs.tell()); //should be 3 * 7 + 3100 = 3121
      fs.close(fd);
    }

    //read file
    Format.printf("Try to open file %s to display\n",A);
    Format.printf("should be 100 lines of 30 chars + newline followed\n");
    Format.printf("by 3 lines ABCDEF + newline followed by EOF\n");
    fd = fs.open(A,fs.READ);
    if (fd <= 0) printError(fd);
    else {
      while ((i=fs.readByte(fd)) > 0) Format.printf("%c",i);
      printError(i); //should be end of file
      fs.close(fd);
    }

    //random access
    Format.printf("Try to random access file %s\n",A);
    fd = fs.open(A,fs.RANDOM);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("changing position 4 to ABCDEFABCDEF\n");
      fs.seek(fd,0,4,fs.SEEK_SET);
      for (i=0; i<2; i++) {    // overwrite pos 4 with "ABCDEFABCDEF"
        for (j=0; j<6; j++) {
          fs.writeByte(fd,'A'+j);
        }
      }
      fs.seek(fd,0,0,fs.SEEK_END);
      fs.tell(fd); //get high word of filesize (=0)
      Format.printf("filesize: %d\n",fs.tell()); //should be 3 * 7 + 3100 = 3121
      fs.close(fd);
    }

    //read file
    Format.printf("Try to open file %s to display\n",A);
    Format.printf("should be ABCDEFABCDEF at position 4\n");
    fd = fs.open(A,fs.READ);
    if (fd <= 0) printError(fd);
    else {
      while ((i=fs.readByte(fd)) > 0) Format.printf("%c",i);
      printError(i); //should be end of file
      fs.close(fd);
    }

    //seek file
    Format.printf("Try to seek before begin and beyond end of file\n");
    fd = fs.open(A,fs.RANDOM);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("move to before begin of file\n");
      error = fs.seek(fd,-1,-1,fs.SEEK_SET); //try to move before begin of file
      printError(error); //should be out of range
      Format.printf("move to beyond end of file\n");
      error = fs.seek(fd,0,1,fs.SEEK_END); //try to move beyond end of file
      printError(error); //should be out of range
      Format.printf("move to middle of file\n");
      error = fs.seek(fd,0,1560,fs.SEEK_SET); //move to middle of file (1560 == 3121/2)
      if (error < 0) printError(error);
      else {
        Format.printf("move to before begin of file from midpoint\n");
        error = fs.seek(fd,-1,-1561,fs.SEEK_CUR); //move back 1561 positions, should fail (1560-1561 == -1)
        printError(error); //should be out of range
        Format.printf("move to beyond end of file from midpoint\n");
        error = fs.seek(fd,0,+1562,fs.SEEK_CUR); //move forward 1562 positions, should fail (1560+1562 == 3122)
        printError(error); //should be out of range
        //character at pos 1561: 1562 characters equals 50*31 + 12
        //this 12 equals 0123456789 + newline + 0
        //so character at 1561 should be 0
        Format.printf("Character at midpoint (pos 1561) is 0\n");
        error = fs.readByte(fd);
        if (error <= 0) printError(error);
        else Format.printf("Character read from file is %c\n",error); //should be 0
      }
      fs.close(fd);
    }

    //truncate file
    Format.printf("Try to truncate file to 2300 bytes\n");
    Format.printf("Open file\n");
    fd = fs.open(A,fs.RANDOM);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("seek to position 2300\n");
      fs.seek(fd,0,2300,fs.SEEK_SET);
      Format.printf("truncate file\n");
      error = fs.truncate(fd);
      if (error < 0) printError(error);
      else {
        Format.printf("close file and open again\n");
        fs.close(fd);
        fd = fs.open(A,fs.READ);
        if (fd <= 0) printError(fd);
        else {
          fs.seek(fd,0,0,fs.SEEK_END);
          fs.tell(fd);
          Format.printf("filesize is %d bytes\n",fs.tell()); //should be 2300
          fs.close(fd);
        }
      }
    }

    //read file
    Format.printf("Try to open file %s to display\n",A);
    Format.printf("should be 2300 bytes\n");
    fd = fs.open(A,fs.READ);
    if (fd <= 0) printError(fd);
    else {
      while ((i=fs.readByte(fd)) > 0) Format.printf("%c",i);
      printError(i); //should be end of file
      fs.close(fd);
    }
*/
/*    //create file B and give list of files
    fd = fs.open(B,fs.WRITE);
    if (fd <= 0) printError(fd);
    else {
      fs.close(fd);
      Format.printf("Listing of files\n");
      j = fs.files();
      i=0;
      while (i < j) {
        Format.printf("%s\n",fs.name(i));
        i++;
      }
    }

    //delete files
    Format.printf("Try to delete files\n");
    error = fs.delete(A);
    if (error < 0) printError(error);
    else {
      Format.printf("File %s deleted\n",A);
      Format.printf("Try to open the file\n");
      fd = fs.open(A,fs.READ);
      printError(fd); //should be file not found
    }
    error = fs.delete(B);
    if (error < 0) printError(error);
    else {
      Format.printf("File %s deleted\n",B);
      Format.printf("Try to open the file\n");
      fd = fs.open(B,fs.READ);
      printError(fd); //should be file not found
    }

*/
//    Format.printf("System ram free: %u bytes\n",stamp.core.Memory.freeMemory());
//    Format.printf("program finished\n");
    System.out.print("program finished\n");
    while (true) ;
  }

  static void printError(int error) {
    switch(error) {
      case err.OK:
        break;
      case err.BAD_FORMAT:
        Format.printf("Bad formatted disk\n");
        break;
      case err.OUT_OF_MEMORY:
        Format.printf("out of memory\n");
        break;
      case err.INTEGRITY_LOST:
        Format.printf("integrity lost\n");
        break;
      case err.INVALID:
        Format.printf("invalid name or descriptor\n");
        break;
      case err.OUT_OF_RANGE:
        Format.printf("out of range\n");
        break;
      case err.NOT_FOUND:
        Format.printf("file not found\n");
        break;
      case err.OPEN:
        Format.printf("file open\n");
        break;
      case err.WRITE_ONLY:
        Format.printf("file write only\n");
        break;
      case err.READ_ONLY:
        Format.printf("file read only\n");
        break;
      case err.NOT_AVAILABLE:
        Format.printf("no file descriptor\n");
        break;
      case err.END:
        Format.printf("end of file\n");
        break;
      case err.FULL:
        Format.printf("disk full\n");
        break;
      default:
        Format.printf("unsupported error %d\n",error);
    }
  }

}