package stamp.filesystem;
import stamp.core.*;
import stamp.errors.*;

/**
 * This class defines a filesystem. It does all the housekeeping
 * regarding file buffering and the reading and writing of file sectors.
 * Its main purpose is to have an easy way of inputting user defined text/data
 * and writing log/data files. It requires a FSinterface object to do the underlying
 * hardware I/O. This FSinterface accesses real physical storage media (aka disk).
 *
 * This class tests wether this device is ready and also checks parameters for proper
 * values before calling an extended class method. So these need not to be checked
 * again in the extended class.
 *
 * This class supports filesizes up to 16MB (with 512 byte sectors) due to the
 * 16 bits signed integers. One such integer is used to hold the logical filesector.
 * 32768 * 512 bytes = 16MB. It does not know about the underlying hardware layout.
 * All files are treated as a large array of sectors. This class calls extended class
 * methods to deal with the hardware related issues.
 *
 * This class requires two values that are passed to it via the constructor:
 * the number of simultaneously open files and the sectorsize.
 *
 * This class makes no assumption about filenames, filename formats or
 * paths and subdirectories. These issues are dealed with by the extended classes.
 *
 * The following public file methods are provided:
 *   present     - test if filesystem device is present
 *   ready       - test if filesystem device is ready
 *   status      - filesystem status
 *   format      - format filesystem
 *   create      - create a file
 *   delete      - delete a file
 *   search      - search a file
 *   open        - open a file for reading and/or writing
 *   close       - close a file
 *   readByte    - read a single byte from a file
 *   readArray   - read multiple bytes from a file
 *   readString  - read string from a file
 *   writeByte   - write a single byte to a file
 *   writeArray  - write multiple bytes to a file
 *   writeString - write string to a file
 *   flush       - write the file buffer to disk
 *   seek        - move to another position within the file
 *   tell        - get the current file position (low word and high word)
 *   truncate    - truncate the filesize to the current fileposition
 *
 * The memory requirements depend on the desired simultaneously opened files.
 * Each open file has a sectorbuffer. Total size is OPENFILES * SECTORSIZE
 * which is the bulk of the used memory in this class.
 *
 * <p><pre>
 * Revision History:
 * September 13, 2004 v1.0: Initial release.
 * </pre><p>
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 September 13, 2004
 */

public class FileSystem {

  /*
    Filesize is limited to 32768 sectors.
    For a 32byte sector max filesize is then 1 megabyte.
    For a 512 byte sector max filesize is then 16 megabytes.
    Max filesize hi:lo equals SECTORSIZE/2 : 0
    Max position hi:lo equals SECTORSIZE/2-1 : 0xFFFF
    To prevent sign problems in calculations the position and size
    are encoded as sector:offset where 0 <= sector <= 32767
                                       0 <= offset <= SECTORSIZE-1
    so both words are positive.
  */

  //seek origin options
  public static final int SEEK_SET = 0; //relative from first position
  public static final int SEEK_CUR = 1; //relative from current position
  public static final int SEEK_END = 2; //relative from end of file position

  //open mode options
  public static final int RANDOM = 0; //read and write, starting at pos 0
  public static final int READ   = 1; //read only, starting at pos 0
  public static final int WRITE  = 2; //write only, starting at pos 0
  public static final int APPEND = 3; //write only, starting at pos size

  private int OPENFILES;         //maximum simultaneously open files

  private class openBlock {
    int index       = 0;      //identifies file on physical disk
    int mode        = 0;      //RANDOM, READ, WRITE or APPEND
    int pos_sector  = 0;      //position sector (currently accessed sector)
    int pos_offset  = 0;      //position offset (index in sector buffer)
    int size_sector = 0;      //current size
    int size_offset = 0;      //
    int disk_sectors = 0;     //sectors on disk
    boolean dirty = false;    //buffer status (CLEAN/DIRTY)
    char[] buf    = null;     //sector buffer (set when opened)
    public openBlock() {
    }
  }

  private openBlock[] fdopen; //array of open file descriptors
  private int tellLo;         //internally stored low word of filesize for method tell()
  private FSinterface fsi;    //the hardware I/O object

  /**
   * Constructor for FileSystem.
   *
   * @param fsi FSinterface for hardware I/O
   * @param openFiles Number of simultaneously open files.
   */
  public FileSystem(FSinterface fsi, int openFiles) {
    int i;
    this.fsi = fsi;
    this.OPENFILES = openFiles;
    fdopen = new openBlock[OPENFILES+1];  //define array
    for (i=1; i<=OPENFILES; i++) { //file descriptors have positve non-zero values
      fdopen[i] = new openBlock(); //initialize array
    }
  }

  /**
   * Test if filesystem device present.
   *
   * @return True if filesystem device present.
   */
  public boolean present() {
    return fsi.present();
  }

  /**
   * Test if filesystem device ready.
   *
   * @return True if filesystem device ready.
   */
  public boolean ready() {
    return fsi.ready();
  }

  /**
   * Get filesystem status.
   *
   * @return errorcode.
   *         Returns 0 if filesytem ok.
   */
  public int status() {
    return fsi.status();
  }

  /**
   * Format filesystem.
   * If no files are opened, this will delete all files.
   * If the filesystem integrity is lost, it will try to create a new filesystem.
   * If that fails the error INTEGRITY_LOST is returned, which is a fatal error.
   *
   * @return errorcode.
   *         Returns 0 if filesytem formatted ok.
   */
  public int format() {
    int fd = 1;
    if (!fsi.ready()) return err.NOT_READY;
    //search for open files
    while (fd <= OPENFILES) {
      if (fdopen[fd].index != 0) return err.OPEN;
      fd++;
    }
    return fsi.formatFileSystem();
  }

  /**
   * Get number of files.
   *
   * @return Number of files.
   */
  public int files() {
    return fsi.files();
  }

  /**
   * Get filename.
   *
   * @param index Logical entry for file.
   *              Range is 0 to files()-1.
   * @return Name of file as ASCIIZ or null.
   */
  public char[] name(int index) {
    return fsi.fileName(index);
  }

  /**
   * Create file.
   *
   * @param name Name of the file.
   *             See extended class for filename specifications.
   * @return errorcode.
   */
  public int create(String name) {
    return create(name.toCharArray());
  }
  public int create(char[] name) {
    int fi;
    if (!fsi.ready()) return err.NOT_READY;
    fi = fsi.searchFile(name);
    if (fi > 0) return err.FOUND; //file already exists
    fi = fsi.createFile(name);
    return (fi < 0) ? fi : err.OK;
  }

  /**
   * Delete file.
   * File can only be deleted if not open.
   *
   * @param name Name of the file.
   *             See extended class for filename specifications.
   * @return errorcode.
   */
  public int delete(String name) {
    return delete(name.toCharArray());
  }
  public int delete(char[] name) {
    int i,fi;
    if (!fsi.ready()) return err.NOT_READY;
    fi = fsi.searchFile(name);
    if (fi > 0) { //file found
      for (i=1; i<=OPENFILES; i++) {
        if (fdopen[i].index == fi) return err.OPEN; //file is currently opened
      }
      return fsi.deleteFile(fi); //file exist and is not open, so it can be deleted
    }
    return fi; //errorcde
  }

  /**
   * Open file.
   * It is allowed to open an already opened file.
   *
   * @param name Name of the file.
   *             See extended class for filename specifications.
   * @param mode RANDOM, READ, WRITE or APPEND.
   *             RANDOM allows read and write access to the file.
   *                    If the file does not exist it is created.
   *             READ   only allows read access to the file.
   *                    The file must exist.
   *             WRITE  allows only write access to the file.
   *                    If the file exists it is deleted, then created again.
   *                    If the file does not exist it is created.
   *             APPEND allows only write access to the file.
   *                    If the file exists the file position is set to end of file.
   *                    If the file does not exist it is created.
   * @return File descriptor or errorcode.
   */
  public int open(String name, int mode) {
    return open(name.toCharArray(),mode);
  }
  public int open(char[] name, int mode) {
    int fd,fi,size_lo,size_hi, error = err.OK;
    if (!fsi.ready()) return err.NOT_READY;
    mode %= 4; //failsafe
    fd = 1;
    //search for free file descriptor
    while (fd <= OPENFILES) {
      if (fdopen[fd].index == 0) break; //found
      fd++;
    }
    if (fd > OPENFILES) return err.NOT_AVAILABLE; //no file descriptor available
    //allocate sector buffer
    if (fdopen[fd].buf == null) {
      if (stamp.core.Memory.freeMemory() >= (fsi.SECTORSIZE+32)) //reserve some overhead
        fdopen[fd].buf = new char[fsi.SECTORSIZE]; //set buffer once
      else return err.OUT_OF_MEMORY; //unable to allocate buffer at this time
    }
    if ((fi = fsi.searchFile(name)) <= 0) { //file not found
      if (mode == READ) return err.NOT_FOUND;
      else {
        fi = fsi.createFile(name);
        if (fi <= 0) return fi; //error
      }
    }
    fdopen[fd].index = fi;
    fdopen[fd].mode = mode;
    fdopen[fd].pos_sector = 0;
    fdopen[fd].pos_offset = 0;
    size_hi = fsi.filesize(fi);
    size_lo = fsi.filesize();
    fdopen[fd].size_sector = fsi.toSector(size_hi,size_lo);
    fdopen[fd].size_offset = fsi.toOffset(size_hi,size_lo);
    fdopen[fd].disk_sectors = fileSectors(fd);
    fdopen[fd].dirty = false;
    if (mode == APPEND) {
      fdopen[fd].pos_sector = fdopen[fd].size_sector;
      fdopen[fd].pos_offset = fdopen[fd].size_offset;
      if (fdopen[fd].pos_offset != 0) error = loadBuffer(fd);
    }
    else { //RANDOM,READ,WRITE
      if ((fdopen[fd].size_sector != 0) || (fdopen[fd].size_offset != 0)) {
        if (mode == WRITE) truncate(fd); //remove any bytes from file
        else error = loadBuffer(fd);
      }
    }
    return (error < 0) ? error : fd;
  }

  /**
   * Close a file.
   * The file will be closed, even if flushing the file buffer failed.
   *
   * @param fd File descriptor.
   * @return errorcode.
   */
  public int close(int fd) {
    int fi,size_lo,size_hi, error = err.OK;
    if (!fsi.ready()) return err.NOT_READY;
    if (!isValid(fd)) return err.INVALID;
    fi = fdopen[fd].index;
    if (fdopen[fd].dirty) error = flushBuffer(fd);
    if (error != err.OK) { //truncate filesize if adding new sector failed
      if (fdopen[fd].pos_sector+1 > fdopen[fd].size_sector) { //must have added new sector
        fdopen[fd].size_offset = 0;
      }
    }
    size_hi = fsi.toHi(fdopen[fd].size_sector,fdopen[fd].size_offset);
    size_lo = fsi.toLo(fdopen[fd].size_sector,fdopen[fd].size_offset);
    fdopen[fd].index = 0;   //remove file from open list
    return (fdopen[fd].mode == READ) ? err.OK : fsi.closeFile(fi,size_hi,size_lo);
  }

  /**
   * Read single byte from file.
   * It is not allowed to read beyond end of file.
   *
   * @param fd File descriptor.
   * @return Value read or errorcode.
   */
  public int readByte(int fd) {
    int data,avail,next,error;
    if (!fsi.ready()) return err.NOT_READY;
    if (!isValid(fd)) return err.INVALID;
    if (fdopen[fd].mode >= WRITE) return err.WRITE_ONLY;
    if (eof(fd)) return err.END;
    data = fdopen[fd].buf[fdopen[fd].pos_offset]&0xFF;
    error = updateFileBufferRead(fd,1);
    return (error < 0) ? error : data;
  }

  /**
   * Read multiple bytes from file.
   * It is not allowed to read beyond end of file.
   *
   * @param fd File descriptor.
   * @param buf Character array to hold characters read.
   * @param num Number of bytes to read.
   * @param offset Start position within buf.
   * @return Number of bytes actually read or errorcode.
   */
  public int readArray(int fd, char[] buf, int num, int offset) {
    int i,start,avail_sec,avail_off,avail,count,done,error=0;
    for (i=0; i<num; i++) {
      error = readByte(fd);
      if (error >= 0) buf[offset+i] = (char)(error & 0xFF);
    }
    return (error < 0) ? error : num;
  }

  /**
   * Read string from file.
   *
   * @param fd File descriptor.
   * @param buf StringBuffer to hold bytes.
   * @param num Number of bytes to read.
   * @return Number of bytes read or errorcode.
   */
  public int readString(int fd, StringBuffer buf, int num) {
    return readArray(fd,buf.toString().toCharArray(),num,0);
  }

  /**
   * Update open file buffer for read.
   * This means updating fileposition, saving the current sector
   * if necessary and loading the next sector.
   *
   * @param fd File descriptor.
   * @param bytes Number of bytes to advance file position.
   * @return errorcode.
   */
  private int updateFileBufferRead(int fd, int bytes) {
    // bytes is limited before call, so offset + bytes <= SECTORSIZE
    int error = err.OK;
    fdopen[fd].pos_offset += bytes; //advance position
    if (fdopen[fd].pos_offset == fsi.SECTORSIZE) { //buffer end reached, must be written to disk if dirty
      fdopen[fd].pos_offset = 0; //sector read/write does not use offset
      if (fdopen[fd].dirty) error = flushBuffer(fd);
      if (error == err.OK) {
        fdopen[fd].pos_sector++;
        if (!eof(fd)) {
          //only load sector if not at eof
          error = loadBuffer(fd);
        }
      }
    }
    return error;
  }

  /**
   * Return distance between new and old position.
   *
   * @param newsec Sector of new position
   * @param newoff Offset of new position
   * @param oldsec Sector of new position
   * @param oldoff Offset of new position
   * @return 32767 if 32767 or more bytes available
   *         0-32766 if less than 32767 bytes available
   */
  private int distance(int newsec, int newoff, int oldsec, int oldoff) {
    int avail_sec,avail_off,avail;
    avail_sec = newsec - oldsec;
    avail_off = newoff - oldoff;
    if (avail_off < 0) {
      avail_sec--;
      avail_off += fsi.SECTORSIZE;
    }
    avail = (avail_sec * fsi.SECTORSIZE) + avail_off;
    if (CPU.carry() || avail<0) return 0x7FFF; //more than 32767 so return max positive
    return avail;
  }

  /**
   * Test if position is end of file.
   *
   * @param fd File descriptor
   * @return True if at end of file position
   */
  private boolean eof(int fd) {
    return (fdopen[fd].pos_sector == fdopen[fd].size_sector) && (fdopen[fd].pos_offset == fdopen[fd].size_offset);
  }

  /**
   * Write single byte to file.
   * When writing to end of file, the filesize will be increased.
   *
   * @param fd File descriptor.
   * @param data Byte to write.
   * @return errorcode.
   */
  public int writeByte(int fd, int data) {
    int error;
    if (!fsi.ready()) return err.NOT_READY;
    if (!isValid(fd)) return err.INVALID;
    if (fdopen[fd].mode == READ) return err.READ_ONLY;
    if (fdopen[fd].pos_sector == fsi.SECTORSIZE/2) return err.OUT_OF_MEMORY; //max filesize
    if (fdopen[fd].pos_offset == fsi.SECTORSIZE) return err.FULL; //disk full
    fdopen[fd].buf[fdopen[fd].pos_offset] = (char)data;
    fdopen[fd].dirty = true;
    return updateFileBufferWrite(fd,1);
  }

  /**
   * Write multiple bytes to file.
   * When writing to end of file, the filesize will be increased.
   *
   * @param fd File descriptor.
   * @param buf Character array holding characters to write.
   * @param num Number of bytes to write.
   * @param offset Start position within buf.
   * @return Number of bytes actually written or errorcode.
   */
  public int writeArray(int fd, char[] buf, int num, int offset) {
    int i,start,avail,count,done,error=0;
    for (i=0; i<num; i++) {
      error = writeByte(fd,buf[offset+i]);
    }
    return (error < 0) ? error : num;
  }

  /**
   * Write string to file.
   *
   * @param fd File descriptor.
   * @param buf String holding bytes.
   */
  public int writeString(int fd, String buf) {
    return writeArray(fd,buf.toCharArray(),buf.length(),0);
  }

  /**
   * Write string to file.
   *
   * @param fd File descriptor.
   * @param buf StringBuffer holding bytes.
   */
  public int writeString(int fd, StringBuffer buf) {
    return writeArray(fd,buf.toString().toCharArray(),buf.length(),0);
  }

  /**
   * Update open file buffer for write.
   * This means updating fileposition, saving the current sector
   * if necessary and loading the next sector.
   *
   * @param fd File descriptor.
   * @param bytes Number of bytes to advance file position.
   * @return errorcode.
   */
  private int updateFileBufferWrite(int fd, int bytes) {
    // bytes is limited before call, so offset + bytes <= SECTORSIZE
    int error = err.OK;
    //advance position
    fdopen[fd].pos_offset += bytes;
    if (fdopen[fd].pos_offset == fsi.SECTORSIZE) { //buffer full, must be written to disk
      fdopen[fd].pos_sector++;
      fdopen[fd].pos_offset = 0;
    }
    //update size
    if (ucmp(fd) > 0) {
      fdopen[fd].size_sector = fdopen[fd].pos_sector;
      fdopen[fd].size_offset = fdopen[fd].pos_offset;
    }
    //flush buffer
    if (fdopen[fd].pos_offset == 0) {
      error = flushBuffer(fd);
      if (error == err.OK) {
        //only load next sector if pos < size
        if (ucmp(fd) < 0) error = loadBuffer(fd);
      }
    }
    return error;
  }

  /**
   * Compare position to size.
   *
   * @param fd File descriptor
   * @return -1 if pos < size
   *          0 if pos == size
   *         +1 if pos > size
   */
  private int ucmp(int fd) {
    int r = fdopen[fd].pos_sector - fdopen[fd].size_sector;
    if (r == 0) r = fdopen[fd].pos_offset - fdopen[fd].size_offset;
    if (r == 0) return 0;
    return (r < 0) ? -1 : 1;
  }

  /**
   * Calculate required sectors for file.
   *
   * @param fd File descriptor
   * @return Number of sectors
   */
  private int fileSectors(int fd) {
    int sectors = fdopen[fd].size_sector;
    if (fdopen[fd].size_offset != 0) sectors++;
    return sectors;
  }

  /**
   * Flush open file buffer.
   *
   * @param fd Fle descriptor.
   * @return errorcode.
   */
  public int flush(int fd) {
    if (!fsi.ready()) return err.NOT_READY;
    if (!isValid(fd)) return err.INVALID;
    int error = flushBuffer(fd);
    if (error == err.OK) {
      int size_hi = fsi.toHi(fdopen[fd].size_sector,fdopen[fd].size_offset);
      int size_lo = fsi.toLo(fdopen[fd].size_sector,fdopen[fd].size_offset);
      fsi.closeFile(fdopen[fd].index,size_hi,size_lo);
    }
    return error;
  }

  /**
   * Flush buffer.
   * This method is the main flush code.
   * It tries to add a file sector if necessary.
   * It does not change size or position.
   *
   * @param fd File descriptor.
   * @return errorcode
   */
  private int flushBuffer(int fd) {
    int error = err.OK;
    if (!fdopen[fd].dirty) return err.OK;
    //we must add a sector when sectors required for size > disk sectors
    int r = fileSectors(fd) - fdopen[fd].disk_sectors;
    if (r > 0) { //must add new sector to file
      error = fsi.addSector(fdopen[fd].index);
      if (error == err.OK) fdopen[fd].disk_sectors++;
    }
    if (error == err.OK) {
      int sector = fdopen[fd].pos_sector;
      if (fdopen[fd].pos_offset == 0) sector--;
      error = fsi.writeSector(fdopen[fd].buf,fdopen[fd].index,sector);
      if (error == err.OK) {
        fdopen[fd].dirty = false;
        //fdopen[fd].pos_offset %= fsi.SECTORSIZE; //make sure offset < SECTORSIZE
      }
    }
    if (error == err.OK) {
      if (r > 0) {
        int size_hi = fsi.toHi(fdopen[fd].size_sector,fdopen[fd].size_offset);
        int size_lo = fsi.toLo(fdopen[fd].size_sector,fdopen[fd].size_offset);
        fsi.closeFile(fdopen[fd].index,size_hi,size_lo);
      }
    }
    return error;
  }

  /**
   * Load buffer.
   * The file descriptor is valid.
   *
   * @param fd File descriptor.
   * @return errorcode.
   */
  private int loadBuffer(int fd) {
    int error = fsi.readSector(fdopen[fd].buf,fdopen[fd].index,fdopen[fd].pos_sector);
    if (error == err.OK) {
      fdopen[fd].dirty = false;
      fdopen[fd].pos_offset %= fsi.SECTORSIZE; //make sure offset < SECTORSIZE
    }
    return error;
  }

  /**
   * Seek to new position.
   * It is an error to seek before the beginning of the file
   * or beyond the end of the file.
   *
   * @param fd File descriptor.
   * @param offset_hi High word of number of byte positions to move relative to origin.
   * @param offset_lo Low word of number of byte positions to move relative to origin.
   * @param origin SEEK_SET, SEEK_CUR, SEEK_END.
   * @return errorcode.
   */
  public int seek(int fd, int offset_hi, int offset_lo, int origin) {
    int newpos_lo, newpos_hi, newpos_sec, newpos_off, cursec, error = err.OK;
    if (!fsi.ready()) return err.NOT_READY;
    if (!isValid(fd)) return err.INVALID;
    newpos_hi = offset_hi;
    newpos_lo = offset_lo;
    switch (origin) {
      case SEEK_SET: break;
      case SEEK_CUR: newpos_lo += fsi.toLo(fdopen[fd].pos_sector,fdopen[fd].pos_offset);
                     if (CPU.carry()) newpos_hi++;
                     newpos_hi += fsi.toHi(fdopen[fd].pos_sector,fdopen[fd].pos_offset);
                     break;
      case SEEK_END: newpos_lo += fsi.toLo(fdopen[fd].size_sector,fdopen[fd].size_offset);
                     if (CPU.carry()) newpos_hi++;
                     newpos_hi += fsi.toHi(fdopen[fd].size_sector,fdopen[fd].size_offset);
                     break;
      default:       return err.ILLEGAL;
    }
    //check new position to be in range 0-EOF
    if ((newpos_hi < 0) || (newpos_hi >= fsi.SECTORSIZE/2)) return err.OUT_OF_RANGE;
    newpos_sec = fsi.toSector(newpos_hi,newpos_lo);
    newpos_off = fsi.toOffset(newpos_hi,newpos_lo);
    if (newpos_sec > fdopen[fd].size_sector) return err.OUT_OF_RANGE;
    if (newpos_sec == fdopen[fd].size_sector) {
      if (newpos_off > fdopen[fd].size_offset) return err.OUT_OF_RANGE;
    }
    //check new sector against current sector
    cursec = fdopen[fd].pos_sector;
    if (newpos_sec != cursec) {
      if (fdopen[fd].dirty) error = flushBuffer(fd);
      if (error == err.OK) { //only load new sector if no error
        fdopen[fd].pos_sector = newpos_sec;
        if ((newpos_sec < fdopen[fd].size_sector) || (newpos_off != 0)) {
          //only load sector if not at eof OR at eof but offset != 0
          error = loadBuffer(fd); //pos_sector must be set prior to loadBuffer
        }
        if (error != err.OK) fdopen[fd].pos_sector = cursec; //restore pos_sector
      }
    }
    //only set new position if no error
    if (error == err.OK) {
      fdopen[fd].pos_sector = newpos_sec;
      fdopen[fd].pos_offset = newpos_off;
    }
    return error;
  }

  /**
   * Get current position within file (high word).
   * This method internally stores the low word.
   * This method must be called before tell().
   * Note that the high word is max SECTORSIZE/2 (for maximum filesize).
   *
   * @param fd File descriptor.
   * @return High word of file position or errorcode.
   */
  public int tell(int fd) {
    int tellHi;
    if (!fsi.ready()) return err.NOT_READY;
    if (!isValid(fd)) return err.INVALID;
    tellLo = fsi.toLo(fdopen[fd].pos_sector,fdopen[fd].pos_offset);
    tellHi = fsi.toHi(fdopen[fd].pos_sector,fdopen[fd].pos_offset);
    return tellHi;
  }

  /**
   * Get current position within file (low word).
   * This returns the internally stored low word.
   * The method tell(fd) must be called first.
   *
   * @return Low word of file position.
   */
  public int tell() {
    return tellLo;
  }

  /**
   * Truncate an open file.
   * Sets the filesize to the current position.
   *
   * @param fd File descriptor.
   * @return errorcode.
   */
  public int truncate(int fd) {
    int i,fi,currentSectors,newSectors;
    if (!fsi.ready()) return err.NOT_READY;
    if (!isValid(fd)) return err.INVALID;
    fi = fdopen[fd].index;
    //current sectors in file
    currentSectors = fdopen[fd].size_sector;
    if (fdopen[fd].size_offset != 0) currentSectors++;
    //new sectors in file
    newSectors = fdopen[fd].pos_sector; //always smaller because pos <= size
    if (fdopen[fd].pos_offset != 0) newSectors++;
    //if size sector change, remove sectors
    if (fdopen[fd].pos_sector != fdopen[fd].size_sector) {
      i = currentSectors - newSectors; //number of sectors to remove
      while (i > 0) {
        fsi.removeSector(fi);
        fdopen[fd].disk_sectors--;
        i--;
      }
      fdopen[fd].size_sector = fdopen[fd].pos_sector;
    }
    fdopen[fd].size_offset = fdopen[fd].pos_offset;
    return err.OK;
  }

  /**
   * Test if file descriptor is valid.
   *
   * @param fd File descriptor.
   * @return True if file descriptor is valid, otherwise false.
   */
  private boolean isValid(int fd) {
    if ((fd <= 0) || (fd > OPENFILES)) return false;
    return (fdopen[fd].index != 0);
  }

}