package stamp.filesystem;
import stamp.peripheral.memory.*;
import stamp.errors.*;

/**
 * This class presents a simple disk implementation. It uses the TableList class
 * for dynamic memory allocation and logical file layout. To preserve memory space,
 * code space and execution time, the filenames are limited to 2 characters.
 * Also, there are no subdirectories.
 *
 * This class extends the FSinterface class. As the FileSystem class checks wether
 * the filesytem is ready and for proper parameter values, these checks are not performed
 * by this class. This class however does supply the present(), ready() and status() methods.
 *
 * <p><pre>
 * Revision History:
 * September 13, 2004 v1.0: Initial release.
 * </pre><p>
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 September 13, 2004
 */

public class TableFSinterface extends FSinterface {

  /*
    Logical filesystem layout:

      root
        |
        |
    fileBlock -- dataBlock -- dataBlock -- etc.
        |
        |
    fileBlock -- dataBlock -- dataBlock -- etc.
        |
        |
    fileBlock -- dataBlock -- dataBlock -- etc.
        |
        |
       etc.

    root (type list), integer value as returned by TableList method listCreate().

    fileBlock (type listnode)
      int name;              //name of file (2 chars)
      int size;              //filesize in bytes
      int data;              //type list, list of dataBlocks

    dataBlock:
      char data[SECTORSIZE]; //storage area data bytes
  */

  //fileBlock fields
  private static final int fbNAME    = 0; //byte offset, filename (2 chars)
  private static final int fbSIZE    = 2; //byte offset, filesize
  private static final int fbDATA    = 4; //byte offset, datalist
  private static final int FILEBLOCK = 6; //bytesize of fileBlock

  //dataBlock fields
  private static final int dbDATA    = 0;  //byte offset, databytes

  private TableList table;
  private int entry;
  private int root;
  private Block blk = new Block(null,0,0);
  private int sizeLo; //stored low word for filesize
  private char[] name = new char[3];

  /**
   * Constructor for TableFSinterface.
   *
   * @param table This is where the filesystem resides.
   * @param entry Entry that holds the list descriptor identifying the filesystem.
   * @param sectorsize Number of bytes in sector (must be power of 2).
   */
  public TableFSinterface(TableList table, int entry, int sectorsize) {
    super(sectorsize);
    this.table = table;
    this.entry = entry;
    this.root = table.listReadEntry(entry);
    this.name[2] = 0; //asciiz
  }

  /**
   * Test if filesystem device present.
   *
   * @return True if filesystem device present.
   */
  protected boolean present() {
    return table.present();
  }

  /**
   * Test if filesystem device ready.
   *
   * @return True if filesystem device ready.
   */
  protected boolean ready() {
    return table.ready();
  }

  /**
   * Reset device.
   * If the device does not support this it must return true.
   *
   * @return True if reset succesful.
   */
  protected boolean reset() {
    return true;
  }

  /**
   * Get filesystem status.
   *
   * @return errorcode.
   *         Returns 0 if filesytem ok.
   */
  protected int status() {
    if (!ready()) return err.NOT_READY;
    return (table.listIntegrity(root)) ? err.OK : err.INTEGRITY_LOST;
  }

  /**
   * Get number of files.
   *
   * @return Number of files.
   */
  protected int files() {
    return table.listItems(root);
  }

  /**
   * Get filename.
   *
   * @param index Logical entry for file.
   *              Range is 0 to files()-1.
   * @return Name of file as ASCIIZ string or null if index is invalid.
   */
  protected char[] fileName(int index) {
    int fi,fname;
    fi = table.listIndex(root,index);
    if (fi > 0) {
      table.listSetBlock(blk,fi);
      fname = blk.readWord(fbNAME);
      name[0] = (char)(fname&0xFF);
      name[1] = (char)(fname>>>8);
      return name;
    }
    return null;
  }

  /**
   * Search file.
   *
   * @param name Name of the file.
   *             Length of name must be at least 2 characters.
   * @return File index or errorcode.
   */
  protected int searchFile(char[] name) {
    int fi,fname;
    fname = (name[0]&0xFF)|(name[1]<<8);
    fi = table.listFirst(root);
    while (fi > 0) {
      table.listSetBlock(blk,fi);
      if (fname == blk.readWord(fbNAME)) return fi; //file found
      fi = table.listNext(fi);
    }
    return err.NOT_FOUND;
  }

  /**
   * Create a new file.
   * Search is done and file does not exist
   *
   * @param name Name of the file.
   * @return File index or errorcode.
   */
  protected int createFile(char[] name) {
    int error,fname;
    int fi = table.listAlloc(FILEBLOCK);
    if (fi <= 0) return fi;
    //fileblock allocated
    fname = (name[0]&0xFF)|(name[1]<<8);
    table.listSetBlock(blk,fi); //access fileblock
    blk.writeWord(fbNAME,fname);
    blk.writeWord(fbSIZE,0);
    blk.writeWord(fbDATA,0);
    table.listAdd(root,fi); //add fileblock to filesystem
    return fi;
  }

  /**
   * Delete a file.
   * The file index is valid.
   *
   * @param fi File index.
   * @return errorcode.
   */
  protected int deleteFile(int fi) {
    table.listSetBlock(blk,fi);
    int fdata = blk.readWord(fbDATA);
    if (fdata > 0) {
      table.listClear(fdata);  //free datablocks (listnodes)
      table.listDelete(fdata); //free data (list)
    }
    table.listRemove(root,fi); //remove fileblock from filesystem
    table.listFree(fi);        //free fileblock
    return err.OK;
  }

  /**
   * Close a file.
   * The new size will be written to disk
   * The file index is valid. Because filesizes in TableFileSystem
   * are only 16bits, parameter size_hi is ignored.
   *
   * @param fi File index of file to delete.
   * @param size_hi High word of new filesize
   * @param size_lo Low word of new filesize
   * @return errorcode.
   */
  protected int closeFile(int fi, int size_hi, int size_lo) {
    table.listSetBlock(blk,fi);
    blk.writeWord(fbSIZE,size_lo);
    return err.OK;
  }

  /**
   * Format filesystem.
   * This will delete all files.
   *
   * @return errorcode.
   *         Returns 0 if filesytem formatted correctly.
   */
  protected int formatFileSystem() {
    int fi,fdata,lnode;
    //free all dataBlocks and fileBlocks but keep root
    if (!table.listIntegrity(root)) {
      root = table.listCreate();
      table.listWriteEntry(entry,root); //save new root
    }
    else {
      //delete files, starting with last
      fi = table.listLast(root);
      while (fi > 0) {
        deleteFile(fi);
        fi = table.listLast(root);
      }
      if (status() != err.OK) { //create new filesystem
        table.listFree(root);
        root = table.listCreate();
        if (root <= 0) {
          root = 0;
          return err.INTEGRITY_LOST; //FATAL ERROR, filesystem lost !!
        }
        table.listWriteEntry(entry,root); //save new root
      }
    }
    return err.OK; //formatting ok
  }

  /**
   * Get file size (high word).
   * This method internally stores the low word.
   * The file index is valid.
   *
   * @param fi File index.
   * @return High word of filesize or errorcode.
   */
  protected int filesize(int fi) {
    table.listSetBlock(blk,fi);
    sizeLo = blk.readWord(fbSIZE);
    return 0; //high word always 0 for TableFileSystem
  }

  /**
   * Get file size (low word).
   * This returns the internally stored low word.
   *
   * @return Low word of filesize.
   */
  protected int filesize() {
    return sizeLo;
  }

  /**
   * Write file sector.
   * This method writes the transfer buffer to a physical sector.
   * The file index is valid.
   *
   * @param buf Sector buffer.
   * @param fi File index.
   * @param sector Logical file sector.
   * @return errorcode.
   */
  protected int writeSector(char[] buf, int fi, int sector) {
    table.listSetBlock(blk,fi);
    int fdata = blk.readWord(fbDATA); //get datalist
    if (fdata > 0) {
      sector = table.listIndex(fdata,sector); //get logical file sector
      if (sector > 0) {
        table.listSetBlock(blk,sector);
        blk.writeArray(0,buf,SECTORSIZE,0);
        return err.OK;
      }
    }
    return err.NOT_FOUND;
  }

  /**
   * Read file sector.
   * This method reads a physical sector into the transfer buffer.
   * The file index is valid.
   *
   * @param buf Sector buffer.
   * @param fi File index.
   * @param sector Logical file sector.
   * @return errorcode.
   */
  protected int readSector(char[] buf, int fi, int sector) {
    table.listSetBlock(blk,fi);
    int fdata = blk.readWord(fbDATA); //get datalist
    if (fdata > 0) {
      sector = table.listIndex(fdata,sector); //get logical file sector
      if (sector > 0) {
        table.listSetBlock(blk,sector);
        blk.readArray(0,buf,SECTORSIZE,0);
        return err.OK;
      }
    }
    return err.NOT_FOUND;
  }

  /**
   * Add sector to file.
   * The file index is valid.
   *
   * @param fi File index.
   * @return errorcode.
   */
  protected int addSector(int fi) {
    int fdata,lnode;
    table.listSetBlock(blk,fi);
    fdata = blk.readWord(fbDATA);
    if (fdata == 0) {
      //create datalist
      fdata = table.listCreate();
      if (fdata <= 0) return fdata; //error
      blk.writeWord(fbDATA,fdata);
    }
    lnode = table.listAlloc(SECTORSIZE);
    if (lnode <= 0) return lnode; //error
    table.listAdd(fdata,lnode);
    return err.OK;
  }

  /**
   * Remove sector from file.
   * The file index is valid.
   *
   * @param fi File index.
   * @return errorcode.
   */
  protected int removeSector(int fi) {
    int fdata,lnode;
    table.listSetBlock(blk,fi);
    fdata = blk.readWord(fbDATA);
    if (fdata > 0) {
      lnode = table.listLast(fdata);
      if (lnode > 0) {
        table.listRemove(fdata,lnode);
        table.listFree(lnode);           //remove datablock
      }
      if (table.listItems(fdata) == 0) { //list empty
        table.listDelete(fdata);         //remove listheader
        blk.writeWord(fbDATA,0);
      }
    }
    return err.OK;
  }

  /**
   * Test if file index is valid.
   *
   * @param fi File index as returned by searchFile().
   * @return True if fi valid, false otherwise.
   */
  private boolean isValid(int fi) {
    int fc; //running fi to compare to fi.
    if (fi <= 0) return false; //file index is always positive non-zero.
    fc = table.listFirst(root);
    while (fc > 0) {
      if (fc == fi) return true;
      fc = table.listNext(fc);
    }
    return false;
  }

}