import stamp.peripheral.memory.*;
import stamp.util.text.*;
import stamp.errors.*;
import stamp.filesystem.*;
import stamp.core.*;

/**
 * This class tests the FileSystem class with a TableFSinterface object.
 * In fact, this is a test for all memory objects: Memory, Block, TableHeap
 * and TableList.
 * The TableFSinterface is implemented using the TableList object. The TableList is
 * implemented using the TableHeap object. The TableHeap is implemented using the
 * Memory object.
 * This test creates a filesystem in Javelin sram. It reads and writes files
 * and you observe wether at some point the files get corrupted.
 * Obviously, that should never be.
 *
 * <p><pre>
 * Revision History:
 * September 13, 2004 v1.0: Initial release.
 * </pre><p>
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 September 13, 2004
 */

public class TableFileSystem_test {

  //create ram buffer
  static char[] ramheapbuf = new char[1500];
  //convert ram buffer to memory object
  static SystemRam myMem = new SystemRam(ramheapbuf,0,1500);
  //create table heap, make 4 entries
  static TableHeap heap = new TableHeap(myMem,4);
  //create table list
  static TableList list = new TableList(heap);
  //create table filesystem interface, store reference in entry 1
  static TableFSinterface fsi = new TableFSinterface(list,1,32);
  //finally, create filesystem
  static FileSystem fs = new FileSystem(fsi,10);

  static final String A = new String("A1");
  static final String B = new String("B2");
  static int root;
  static int file;

  static void main() {
    int i,j,fd,error;
    Format.printf("TableFileSystem test program\n");
    if (!heap.heapIntegrity()) {
      Format.printf("heap integrity lost\n");
      Format.printf("Program halted\n");
      while (true) ;
    }
    if (fs.status() != err.OK) { //try to create filesystem
      if (fs.format() != err.OK) {
        Format.printf("Unable to create filesystem\n");
        Format.printf("Program halted\n");
        while (true) ;
      }
    }
    root = list.listReadEntry(1); //entry 1 was defined to hold filesystem reference
    Format.printf("heap initial size = 1500 bytes\n");
    Format.printf("filesystem root wordaddress = %04x\n",root);

    //create file.
    Format.printf("Try to create file %s\n",A);
    fd = fs.open(A,fs.WRITE);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("writing 7 lines 0123456789 = 77 bytes\n");
      for (i=0; i<7; i++) {    // 7 lines of "0123456789" + newline
        for (j=0; j<10; j++) {
          fs.writeByte(fd,'0'+j);
        }
        fs.writeByte(fd,10); //newline
      }
      fs.seek(fd,0,0,fs.SEEK_END);
      fs.tell(fd); //get high word of filesize (=0)
      Format.printf("filesize: %d\n",fs.tell()); //should be 7 * 11 = 77
      fs.close(fd);
    }

    //read file
    Format.printf("Try to open file %s to display\n",A);
    Format.printf("should be 7 lines 0123456789 followed by EOF\n");
    fd = fs.open(A,fs.READ);
    if (fd <= 0) printError(fd);
    else {
      while ((i=fs.readByte(fd)) > 0) Format.printf("%c",i);
      printError(i); //should be end of file
      fs.close(fd);
    }

    //append to file
    Format.printf("Try to append to file %s\n",A);
    fd = fs.open(A,fs.APPEND);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("appending 3 lines ABCDEF = 21 bytes\n");
      for (i=0; i<3; i++) {    // 3 lines of "ABCDEF" + newline
        for (j=0; j<6; j++) {
          fs.writeByte(fd,'A'+j);
        }
        fs.writeByte(fd,10); //newline
      }
      fs.seek(fd,0,0,fs.SEEK_END);
      fs.tell(fd); //get high word of filesize (=0)
      Format.printf("filesize: %d\n",fs.tell()); //should be 3 * 7 + 77 = 98
      fs.close(fd);
    }

    //read file
    Format.printf("Try to open file %s to display\n",A);
    Format.printf("should be 7 lines 0123456789 followed\n");
    Format.printf("by 3 lines ABCDEF followed by EOF\n");
    fd = fs.open(A,fs.READ);
    if (fd <= 0) printError(fd);
    else {
      while ((i=fs.readByte(fd)) > 0) Format.printf("%c",i);
      printError(i); //should be end of file
      fs.close(fd);
    }

    //random access
    Format.printf("Try to random access file %s\n",A);
    fd = fs.open(A,fs.RANDOM);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("changing position 4 to ABCDEFABCDEF\n");
      fs.seek(fd,0,4,fs.SEEK_SET);
      for (i=0; i<2; i++) {    // overwrite pos 4 with "ABCDEFABCDEF"
        for (j=0; j<6; j++) {
          fs.writeByte(fd,'A'+j);
        }
      }
      fs.seek(fd,0,0,fs.SEEK_END);
      fs.tell(fd); //get high word of filesize (=0)
      Format.printf("filesize: %d\n",fs.tell()); //should be 3 * 7 + 77 = 98
      fs.close(fd);
    }

    //read file
    Format.printf("Try to open file %s to display\n",A);
    Format.printf("should be ABCDEFABCDEF at position 4\n");
    fd = fs.open(A,fs.READ);
    if (fd <= 0) printError(fd);
    else {
      while ((i=fs.readByte(fd)) > 0) Format.printf("%c",i);
      printError(i); //should be end of file
      fs.close(fd);
    }

    //seek file
    Format.printf("Try to seek before begin and beyond end of file\n");
    fd = fs.open(A,fs.RANDOM);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("move to before begin of file\n");
      error = fs.seek(fd,-1,-1,fs.SEEK_SET); //try to move before begin of file
      printError(error); //should be out of range
      Format.printf("move to beyond end of file\n");
      error = fs.seek(fd,0,1,fs.SEEK_END); //try to move beyond end of file
      printError(error); //should be out of range
      Format.printf("move to middle of file\n");
      error = fs.seek(fd,0,49,fs.SEEK_SET); //move to middle of file (49 == 98/2)
      if (error < 0) printError(error);
      else {
        Format.printf("move to before begin of file from midpoint\n");
        error = fs.seek(fd,-1,-50,fs.SEEK_CUR); //move back 50 positions, should fail (49-50 == -1)
        printError(error); //should be out of range
        Format.printf("move to beyond end of file from midpoint\n");
        error = fs.seek(fd,0,+50,fs.SEEK_CUR); //move forward 50 positions, should fail (49+50 == 99)
        printError(error); //should be out of range
        Format.printf("Character at midpoint (pos 49) is 5\n");
        error = fs.readByte(fd);
        if (error <= 0) printError(error);
        else Format.printf("Character read from file is %c\n",error); //should be 5
      }
      fs.close(fd);
    }

    //truncate file
    Format.printf("Try to truncate file to 64 bytes\n");
    Format.printf("Open file\n");
    fd = fs.open(A,fs.RANDOM);
    if (fd <= 0) printError(fd);
    else {
      Format.printf("seek to position 64\n");
      fs.seek(fd,0,64,fs.SEEK_SET);
      Format.printf("truncate file\n");
      error = fs.truncate(fd);
      if (error < 0) printError(error);
      else {
        Format.printf("close file and open again\n");
        fs.close(fd);
        fd = fs.open(A,fs.READ);
        if (fd <= 0) printError(fd);
        else {
          fs.seek(fd,0,0,fs.SEEK_END);
          if ((i=fs.tell(fd)) < 0) printError(i);
          else Format.printf("filesize is %d bytes\n",fs.tell()); //should be 64
          fs.close(fd);
        }
      }
    }

    //read file
    Format.printf("Try to open file %s to display\n",A);
    Format.printf("should be 64 bytes\n");
    fd = fs.open(A,fs.READ);
    if (fd <= 0) printError(fd);
    else {
      while ((i=fs.readByte(fd)) > 0) Format.printf("%c",i);
      printError(i); //should be end of file
      fs.close(fd);
    }

    //create file B and give list of files
    fd = fs.open(B,fs.WRITE);
    if (fd <= 0) printError(fd);
    else {
      fs.close(fd);
      Format.printf("Listing of files\n");
      j = fs.files();
      i=0;
      while (i < j) {
        Format.printf("%s\n",fs.name(i));
        i++;
      }
    }

    //delete files
    Format.printf("Try to delete files\n");
    error = fs.delete(A);
    if (error < 0) printError(error);
    else {
      Format.printf("File %s deleted\n",A);
      Format.printf("Try to open the file\n");
      fd = fs.open(A,fs.READ);
      printError(fd); //should be file not found
    }
    error = fs.delete(B);
    if (error < 0) printError(error);
    else {
      Format.printf("File %s deleted\n",B);
      Format.printf("Try to open the file\n");
      fd = fs.open(B,fs.READ);
      printError(fd); //should be file not found
    }

    Format.printf("Heap free: %d bytes\n",heap.heapAvailable());
    Format.printf("System ram free: %d bytes\n",stamp.core.Memory.freeMemory());
    Format.printf("program finished\n");
    while (true) ;
  }

  static void printError(int error) {
    switch(error) {
      case err.OK:
        break;
      case err.OUT_OF_MEMORY:
        Format.printf("out of memory\n");
        break;
      case err.INTEGRITY_LOST:
        Format.printf("integrity lost\n");
        break;
      case err.INVALID:
        Format.printf("invalid name or descriptor\n");
        break;
      case err.OUT_OF_RANGE:
        Format.printf("out of range\n");
        break;
      case err.NOT_FOUND:
        Format.printf("file not found\n");
        break;
      case err.OPEN:
        Format.printf("file open\n");
        break;
      case err.WRITE_ONLY:
        Format.printf("file write only\n");
        break;
      case err.READ_ONLY:
        Format.printf("file read only\n");
        break;
      case err.NOT_AVAILABLE:
        Format.printf("no file descriptor\n");
        break;
      case err.END:
        Format.printf("end of file\n");
        break;
      default:
        Format.printf("unsupported error %d\n",error);
    }
  }

}