/*
 * @file DLlist.java
 *
 * Defines an doubly linked list Abstract Data Type
 * Data can be added unsorted or insert sorted
 *
 * Add performance is O(1)    // unsorted
 * Add performance is O(N)    // sorted
 * Delete performance is O(1) // delete by reference
 * Find performance is O(N)   // unsorted
 * Find performance is O(N)   // sorted
 *
 */
public class DLlist
{
  protected DLnode top;       // head of list
  protected DLnode end;       // end of list
  protected DLnode nextnode;  // next in list iterator
  protected DLnode prevnode;  // prev in list iterator
  protected int size;

  public DLlist()
  {
    top  = null;
    end  = null;
    nextnode = top;
    prevnode = top;
    size = 0;
  }


  /*
   * add node by reference to end of list
   */
  public void add(Object obj)
  {
    DLnode node = new DLnode(obj);
    if(top == null) {
      top = node;
      end = node;
    }
    else {
      end.setNext(node);  // old end node points to new node
      node.setPrev(end);  // node prev points to old end node
      end = node;         // new end node is new node
    }
    size++;
  }

/*
  private void addTop(DLnode node)
  {
    node.setNext(top);
    top.setPrev(node);
    top = node;
  }
  private void addEnd(DLnode node)
  {
    node.setPrev(end);
    end.setNext(node);
    end = node;
  }
  private void addBefore(DLnode next, DLnode node)
  {
    next.setNext(node); // next --> node
    node.setPrev(next); // node <-- next
    next = node;        // node is next
  }

  //
   * insert node to list sorted by string value
   //
  public void insert(String string)
  {
    DLnode node = new DLnode(string);
    DLnode next = top;

    if(next == null) {
      // add to empty list
      top = end = node;
    }
    while(next != null) {
      if (next.getNext() == null) {  // last node
        if(string.compareTo((String)next.getObject()) > 0) {
          addEnd(node);
          break;
        }
        else {
          addTop(node);
          break;
        }
      }
      else if(string.compareTo((String)next.getNext().getObject()) <= 0) {
        if(string.compareTo((String)next.getObject()) > 0)
          addBefore(next.getNext(), node);
      }
      next = next.getNext();
    }
    size++;
  }
*/

  /*
   * insert node to list sorted by string value
   */
  public void insert(String string)
  {
    DLnode node = new DLnode(string);
    DLnode next = top;
    DLnode prev = null;

    // add to empty list
    if(next == null) {
      top = end = node;
      //System.out.println("top " + string);
    }

    // add to sorted list
    while(next != null) {
      String nstr = (String)next.getObject();
      //System.out.println(string + " <= " + nstr + "?");
      if(string.compareTo(nstr) <= 0) {
        if(prev == null) {
          //System.out.println("top " + string);
          node.setNext(top);
          top.setPrev(node);
          top = node;
        }
        else {
          //System.out.println("insert " + string);
          node.setNext(next);
          node.setPrev(prev);
          next.setPrev(node);
          prev.setNext(node);
        }
        break;
      }
      prev = next;
      next = next.getNext();

      if (next == null) {  // last node
        //System.out.println("end " + string);
        end.setNext(node);
        node.setPrev(prev);
        end = node;
      }
    }
    size++;
  }

  /*
   * delete node by reference
   */
  public void delete(DLnode obj)
  {
    if(top == null)
      return;

    //System.out.println("delete " + obj.getObject());

    DLnode next = obj.getNext();
    DLnode prev = obj.getPrev();

    if(next == null && prev == null) {
      top = end = null;
    }
    else if(next == null) {
      prev.setNext(null);
      end = prev;
    }
    else if(prev == null) {
      next.setPrev(null);
      top = next;
    }
    else {
      prev.setNext(next);
      next.setPrev(prev);
    }
    size--;
  }

  /*
   * delete node by string value
   */
  public void delete(String string)
  {
    DLnode node = find(string);
    if(node != null)
      delete(node);
  }

  /*
   * find node by value
   */
  public DLnode find(String string)
  {
    if(size == 0)
      return (DLnode)null;

    DLnode next = top;
    while(next != null) {
      Object ob = next.getObject();
      if(string.equals((String)ob))
        return next;
      next = next.getNext();
    }
    return (DLnode)null;
  }

  /*
   * get DLlist length
   */
  public int length()
  {
    return size;
  }

  /*
   * get next if available
   */
  public DLnode hasNext()
  {
    DLnode node = nextnode;
    if(nextnode == null) {
      nextnode = top;
      return null;
    }
    nextnode = nextnode.getNext();
    return node;
  }
  /*
   * get prev if available
   */
  public DLnode hasPrev()
  {
    DLnode node = prevnode;
    if(prevnode == null) {
      prevnode = end;
      return null;
    }
    prevnode = prevnode.getPrev();
    return node;
  }


  /*
   * show list
   */
  public void show()
  {
    System.out.println(" ");
    System.out.println("Found " + length() + " Total Elements:");
    DLnode next = top;
    while(next != null) {
      Object ob = next.getObject();
      System.out.println((String)ob);
      next = next.getNext();
    }
  }
}

